#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2020-2016 Gramps Development Team
# Copyright (C) 2020      Paul Culley
# Copyright (C) 2024      Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""Generic upgrade module for dbapi dbs"""
# ------------------------------------------------------------------------
#
# Python Modules
#
# ------------------------------------------------------------------------
import os
import re
import time
import logging

# ------------------------------------------------------------------------
#
# Gramps Modules
#
# ------------------------------------------------------------------------
from gramps.cli.clidbman import NAME_FILE
from gramps.gen.db.dbconst import CLASS_TO_KEY_MAP
from gramps.gen.lib import EventType, NameOriginType, Tag, MarkerType
from gramps.gen.utils.file import create_checksum
from gramps.gen.utils.id import create_id
from gramps.gui.dialog import InfoDialog
from .dbconst import (
    PERSON_KEY,
    FAMILY_KEY,
    EVENT_KEY,
    MEDIA_KEY,
    PLACE_KEY,
    REPOSITORY_KEY,
    CITATION_KEY,
    SOURCE_KEY,
    NOTE_KEY,
    TAG_KEY,
)
from ..const import GRAMPS_LOCALE as glocale
from .conversion_tools import convert_21

_ = glocale.translation.gettext

LOG = logging.getLogger(".upgrade")


def gramps_upgrade_21(self):
    """
    Add json_data field to tables.
    """
    self.set_serializer("blob")

    length = 0
    for key in self._get_table_func():
        count_func = self._get_table_func(key, "count_func")
        length += count_func()

    self.set_total(length)

    # First, do metadata:

    self._txn_begin()
    self.upgrade_table_for_json_data("metadata")
    # the *map_keys are not in some older versions of the db so we will initialize
    # them here; needed in blobs if someone tries to downgrade the db so we don't
    # crash before the warning is generated.
    self.set_serializer("blob")
    self._set_metadata("cmap_index", self.cmap_index, use_txn=False)
    self._set_metadata("smap_index", self.smap_index, use_txn=False)
    self._set_metadata("emap_index", self.emap_index, use_txn=False)
    self._set_metadata("pmap_index", self.pmap_index, use_txn=False)
    self._set_metadata("fmap_index", self.fmap_index, use_txn=False)
    self._set_metadata("lmap_index", self.lmap_index, use_txn=False)
    self._set_metadata("omap_index", self.omap_index, use_txn=False)
    self._set_metadata("rmap_index", self.rmap_index, use_txn=False)
    self._set_metadata("nmap_index", self.nmap_index, use_txn=False)

    keys = self._get_metadata_keys()
    for key in keys:
        self.set_serializer("blob")
        value = self._get_metadata(key, "not-found")
        if value != "not-found":
            # Save to json_data in current format
            self.set_serializer("json")
            self._set_metadata(key, value, use_txn=False)

    for table_name in self._get_table_func():
        # For each table, alter the database in an appropriate way:
        self.upgrade_table_for_json_data(table_name.lower())

        get_array_from_handle = self._get_table_func(table_name, "raw_func")
        get_object_from_handle = self._get_table_func(table_name, "handle_func")
        get_handles = self._get_table_func(table_name, "handles_func")
        commit_func = self._get_table_func(table_name, "commit_func")
        key = CLASS_TO_KEY_MAP[table_name]
        for handle in get_handles():
            # Load from blob:
            self.set_serializer("blob")
            array = get_array_from_handle(handle)
            # Save to json_data in version 21 format
            json_data = convert_21(table_name, array)
            self.set_serializer("json")
            # We can use commit_raw as long as json_data
            # has "handle", "_class", and uses json.dumps()
            self._commit_raw(json_data, key)
            self.update()
        self._drop_column(table_name, "blob_data")

    self.set_serializer("blob")
    self._set_metadata("version", 21, use_txn=False)  # keep blob version up to date
    self.set_serializer("json")
    self._set_metadata("version", 21, use_txn=False)
    self._txn_commit()
    # Bump up database version. Separate transaction to save metadata.


def gramps_upgrade_20(self):
    """
    Upgrade to database version 20
    """
    self.set_serializer("blob")

    length = 0
    self.set_total(length)
    self._txn_begin()

    # uid and place upgrade code goes here

    # ----------------------------------------
    # Modify Person
    # ----------------------------------------
    # Add citation_list to person eventref objects
    for person_handle in self.get_person_handles():
        person = self.get_raw_person_data(person_handle)
        (
            handle,
            gramps_id,
            gender,
            primary_name,
            alternate_names,
            death_ref_index,
            birth_ref_index,
            event_ref_list,
            family_list,
            parent_family_list,
            media_list,
            address_list,
            attribute_list,
            urls,
            lds_seal_list,
            citation_list,
            note_list,
            change,
            tag_list,
            private,
            person_ref_list,
        ) = person
        if event_ref_list:
            event_ref_list = upgrade_event_ref_list_20(event_ref_list)
            new_person = (
                handle,
                gramps_id,
                gender,
                primary_name,
                alternate_names,
                death_ref_index,
                birth_ref_index,
                event_ref_list,
                family_list,
                parent_family_list,
                media_list,
                address_list,
                attribute_list,
                urls,
                lds_seal_list,
                citation_list,
                note_list,
                change,
                tag_list,
                private,
                person_ref_list,
            )
            self._commit_raw(new_person, PERSON_KEY)
        self.update()

    # ----------------------------------------
    # Modify Family
    # ----------------------------------------
    # Add citation_list to family eventref objects
    for family_handle in self.get_family_handles():
        family = self.get_raw_family_data(family_handle)
        (
            handle,
            gramps_id,
            father_handle,
            mother_handle,
            child_ref_list,
            the_type,
            event_ref_list,
            media_list,
            attribute_list,
            lds_seal_list,
            citation_list,
            note_list,
            change,
            marker,
            private,
        ) = family
        if event_ref_list:
            event_ref_list = upgrade_event_ref_list_20(event_ref_list)
            new_family = (
                handle,
                gramps_id,
                father_handle,
                mother_handle,
                child_ref_list,
                the_type,
                event_ref_list,
                media_list,
                attribute_list,
                lds_seal_list,
                citation_list,
                note_list,
                change,
                marker,
                private,
            )
            self._commit_raw(new_family, FAMILY_KEY)
        self.update()

    self._txn_commit()
    # Bump up database version. Separate transaction to save metadata.
    self._set_metadata("version", 20)


def upgrade_event_ref_list_20(event_ref_list):
    """
    Insert citation_list into eventref objects
    """
    new_event_ref_list = []
    for event_ref in event_ref_list:
        (privacy, note_list, attribute_list, ref, role) = event_ref
        new_event_ref = (privacy, [], note_list, attribute_list, ref, role)
        new_event_ref_list.append((new_event_ref))
    return new_event_ref_list


def gramps_upgrade_19(self):
    """
    Upgrade database from version 18 to 19.
    """
    self.set_serializer("blob")
    # This is done in the conversion from bsddb, so just say we did it.
    self._set_metadata("version", 19)


def gramps_upgrade_18(self):
    """
    Upgrade database from version 17 to 18.
    """
    self.set_serializer("blob")

    length = self.get_number_of_places()
    self.set_total(length)
    self._txn_begin()

    # ---------------------------------
    # Modify Place
    # ---------------------------------
    # Convert name fields to use PlaceName.
    for handle in self.get_place_handles():
        place = self.get_raw_place_data(handle)
        new_place = list(place)
        new_place[6] = (new_place[6], None, "")
        alt_names = []
        for name in new_place[7]:
            alt_names.append((name, None, ""))
        new_place[7] = alt_names
        new_place = tuple(new_place)
        self._commit_raw(new_place, PLACE_KEY)
        self.update()

    self._txn_commit()
    # Bump up database version. Separate transaction to save metadata.
    self._set_metadata("version", 18)


def gramps_upgrade_17(self):
    """
    Upgrade database from version 16 to 17.

    1. This upgrade adds tags to event, place, repository, source and
       citation objects.
    2. Data of Source becomes SourceAttributes Secondary Object
    3. Create a place hierarchy.
    4. Add checksum field to media objects.
    5. Rebuild list of custom events.
    """
    self.set_serializer("blob")

    length = (
        self.get_number_of_events()
        + self.get_number_of_places()
        + self.get_number_of_citations()
        + self.get_number_of_sources()
        + self.get_number_of_repositories()
        + self.get_number_of_media()
    )
    self.set_total(length)
    self._txn_begin()

    # ---------------------------------
    # Modify Event
    # ---------------------------------
    # Add new tag_list field.
    self.event_names = set()
    for handle in self.get_event_handles():
        event = self.get_raw_event_data(handle)
        # Pre version 17 event type:
        new_event = list(event)
        # Version 17 event type:
        event_type = EventType()
        event_type.unserialize(new_event[2])
        if event_type.is_custom():
            self.event_names.add(str(event_type))
        new_event = new_event[:11] + [[]] + new_event[11:]
        new_event = tuple(new_event)
        self._commit_raw(new_event, EVENT_KEY)
        self.update()

    # ---------------------------------
    # Modify Place
    # ---------------------------------
    # Convert to hierarchical structure and add new tag_list field.
    locations = {}
    self.max_id = 0
    index = re.compile("[0-9]+")
    for handle in self.get_place_handles():
        place = self.get_raw_place_data(handle)
        match = index.search(place[1])
        if match:
            if self.max_id <= int(match.group(0)):
                self.max_id = int(match.group(0))
        if place[5] is not None:
            locations[get_location(place[5])] = handle

    for handle in list(self.get_place_handles()):
        place = self.get_raw_place_data(handle)
        new_place = list(place)

        zip_code = ""
        if new_place[5]:
            zip_code = new_place[5][0][6]

        # find title and type
        main_loc = get_location(new_place[5])
        for level, name in enumerate(main_loc):
            if name:
                break

        loc = list(main_loc[:])
        loc[level] = ""

        # find top parent
        parent_handle = None
        for n in range(7):
            if loc[n]:
                tup = tuple([""] * n + loc[n:])
                parent_handle = locations.get(tup, None)
                if parent_handle:
                    break

        # create nodes
        if parent_handle:
            n -= 1
        while n > level:
            if loc[n]:
                # TODO for Arabic, should the next line's comma be translated?
                title = ", ".join([item for item in loc[n:] if item])
                parent_handle = add_place(self, loc[n], n, parent_handle, title)
                locations[tuple([""] * n + loc[n:])] = parent_handle
            n -= 1

        if parent_handle is not None:
            if isinstance(parent_handle, bytes):
                parent_handle = parent_handle.decode("utf-8")
            placeref_list = [(parent_handle, None)]
        else:
            placeref_list = []

        if name:
            type_num = 7 - level
        else:
            name = new_place[2]
            type_num = -1
        new_place = (
            new_place[:5]
            + [placeref_list, name, [], (type_num, ""), zip_code]
            + new_place[6:12]
            + [[]]
            + new_place[12:]
        )
        new_place = tuple(new_place)
        self._commit_raw(new_place, PLACE_KEY)
        self.update()

    # ---------------------------------
    # Modify Repository
    # ---------------------------------
    # Add new tag_list field.
    for handle in self.get_repository_handles():
        repository = self.get_raw_repository_data(handle)
        new_repository = list(repository)
        new_repository = new_repository[:8] + [[]] + new_repository[8:]
        new_repository = tuple(new_repository)
        self._commit_raw(new_repository, REPOSITORY_KEY)
        self.update()

    # ---------------------------------
    # Modify Source
    # ---------------------------------
    # Add new tag_list field.
    for handle in self.get_source_handles():
        source = self.get_raw_source_data(handle)
        new_source = list(source)
        new_source = new_source[:11] + [[]] + new_source[11:]
        new_source = tuple(new_source)
        self._commit_raw(new_source, SOURCE_KEY)
        self.update()

    # ---------------------------------
    # Modify Citation
    # ---------------------------------
    # Add new tag_list field.
    for handle in self.get_citation_handles():
        citation = self.get_raw_citation_data(handle)
        new_citation = list(citation)
        new_citation = new_citation[:10] + [[]] + new_citation[10:]
        new_citation = tuple(new_citation)
        self._commit_raw(new_citation, CITATION_KEY)
        self.update()

    # -------------------------------------------------------
    # Upgrade Source and Citation datamap to SrcAttributeBase
    # -------------------------------------------------------
    for handle in self.get_source_handles():
        source = self.get_raw_source_data(handle)
        (
            handle,
            gramps_id,
            title,
            author,
            pubinfo,
            notelist,
            medialist,
            abbrev,
            change,
            datamap,
            reporef_list,
            taglist,
            private,
        ) = source
        srcattributelist = upgrade_datamap_17(datamap)
        new_source = (
            handle,
            gramps_id,
            title,
            author,
            pubinfo,
            notelist,
            medialist,
            abbrev,
            change,
            srcattributelist,
            reporef_list,
            taglist,
            private,
        )
        self._commit_raw(new_source, SOURCE_KEY)
        self.update()

    for handle in self.get_citation_handles():
        citation = self.get_raw_citation_data(handle)
        (
            handle,
            gramps_id,
            datelist,
            page,
            confidence,
            source_handle,
            notelist,
            medialist,
            datamap,
            change,
            taglist,
            private,
        ) = citation
        srcattributelist = upgrade_datamap_17(datamap)
        new_citation = (
            handle,
            gramps_id,
            datelist,
            page,
            confidence,
            source_handle,
            notelist,
            medialist,
            srcattributelist,
            change,
            taglist,
            private,
        )
        self._commit_raw(new_citation, CITATION_KEY)
        self.update()

    # ---------------------------------
    # Modify Media
    # ---------------------------------
    # Add new checksum field.
    base_path = self._get_metadata("media-path")
    if not base_path:
        # Check that the mediapath is not set to None (bug #7844).
        base_path = ""
    for handle in self.get_media_handles():
        media = self.get_raw_media_data(handle)
        new_media = list(media)
        if os.path.isabs(new_media[2]):
            full_path = new_media[2]
        else:
            full_path = os.path.join(base_path, new_media[2])
        checksum = create_checksum(full_path)
        new_media = new_media[:5] + [checksum] + new_media[5:]
        new_media = tuple(new_media)
        self._commit_raw(new_media, MEDIA_KEY)
        self.update()

    self._txn_commit()
    # Bump up database version. Separate transaction to save metadata.
    self._set_metadata("version", 17)


def get_location(loc):
    # (street, locality, parish, city, county, state, country)
    if loc is None:
        location = ("",) * 7
    else:
        location = loc[0][:2] + (loc[1],) + loc[0][2:6]
    return location


def add_place(self, name, level, parent, title):
    handle = create_id()
    self.max_id += 1
    gid = self.place_prefix % self.max_id
    placetype = (7 - level, "")
    if parent is not None:
        if isinstance(parent, bytes):
            parent = parent.decode("utf-8")
        placeref_list = [(parent, None)]
    else:
        placeref_list = []
    place = (
        handle,
        gid,
        title,
        "",
        "",
        placeref_list,
        name,
        [],
        placetype,
        "",
        [],
        [],
        [],
        [],
        [],
        0,
        [],
        False,
    )
    self._commit_raw(place, PLACE_KEY)
    return handle


def upgrade_datamap_17(datamap):
    """
    In version 16 key value pairs are stored in source and citation.
    These become SrcAttribute
    """
    new_srcattr_list = []
    private = False
    from gramps.gen.lib.srcattrtype import SrcAttributeType

    for key, value in datamap.items():
        the_type = SrcAttributeType(key).serialize()
        new_srcattr_list.append((private, the_type, value))
    return new_srcattr_list


def gramps_upgrade_16(self):
    """
    Upgrade database from version 15 to 16. This upgrade converts all
    SourceRef child objects to Citation Primary objects.

    For each primary object that has a sourceref, what we have to do is:

        (1) create each citation
        (2) update the object to reference the Citations
        (3) remove backlinks for references from object to Source
        (4) add backlinks for references from object to Citations
        (5) add backlinks for references from Citation to Source

    the backlinks are all updated at the end by calling
    :py:meth:`reindex_reference_map <.write.DbBsddb.reindex_reference_map>`
    """
    # Only People, Families, Events, Media Objects, Places, Sources and
    # Repositories need to be updated, because these are the only primary
    # objects that can have source citations.
    self.set_serializer("blob")

    length = (
        self.get_number_of_people()
        + self.get_number_of_events()
        + self.get_number_of_families()
        + self.get_number_of_repositories()
        + self.get_number_of_media()
        + self.get_number_of_places()
        + self.get_number_of_sources()
    )
    self.set_total(length)
    self._txn_begin()

    # Setup data for upgrade statistics information dialogue
    keyorder = [
        PERSON_KEY,
        FAMILY_KEY,
        EVENT_KEY,
        MEDIA_KEY,
        PLACE_KEY,
        REPOSITORY_KEY,
        SOURCE_KEY,
    ]
    key2data = {
        PERSON_KEY: 0,
        FAMILY_KEY: 1,
        EVENT_KEY: 2,
        MEDIA_KEY: 3,
        PLACE_KEY: 4,
        REPOSITORY_KEY: 5,
        SOURCE_KEY: 6,
    }
    key2string = {
        PERSON_KEY: _(
            "%(n1)6d  People        upgraded with "
            "%(n2)6d citations in %(n3)6d secs\n"
        ),
        FAMILY_KEY: _(
            "%(n1)6d  Families      upgraded with "
            "%(n2)6d citations in %(n3)6d secs\n"
        ),
        EVENT_KEY: _(
            "%(n1)6d  Events        upgraded with "
            "%(n2)6d citations in %(n3)6d secs\n"
        ),
        MEDIA_KEY: _(
            "%(n1)6d  Media Objects upgraded with "
            "%(n2)6d citations in %(n3)6d secs\n"
        ),
        PLACE_KEY: _(
            "%(n1)6d  Places        upgraded with "
            "%(n2)6d citations in %(n3)6d secs\n"
        ),
        REPOSITORY_KEY: _(
            "%(n1)6d  Repositories  upgraded with "
            "%(n2)6d citations in %(n3)6d secs\n"
        ),
        SOURCE_KEY: _(
            "%(n1)6d  Sources       upgraded with "
            "%(n2)6d citations in %(n3)6d secs\n"
        ),
    }
    data_upgradeobject = [0] * 7

    # Initialise the citation gramps ID number
    self.cmap_index = 0

    # ---------------------------------
    # Modify Person
    # ---------------------------------
    start_num_citations = self.cmap_index
    start_time = time.time()
    for person_handle in self.get_person_handles():
        person = self.get_raw_person_data(person_handle)
        try:
            # The parameters are evaluated before deciding whether logging is
            # on or not. Since the retrieval of names is so complex, I think it
            # is safer to protect this with a try except block, even though it
            # seems to work for names being present and not.
            LOG.debug(
                "upgrade person %s %s"
                % (person[3][4], " ".join([name[0] for name in person[3][5]]))
            )
        except:
            pass
        (
            handle,
            gramps_id,
            gender,
            primary_name,
            alternate_names,
            death_ref_index,
            birth_ref_index,
            event_ref_list,
            family_list,
            parent_family_list,
            media_list,
            address_list,
            attribute_list,
            urls,
            lds_seal_list,
            source_list,
            note_list,
            change,
            tag_list,
            private,
            person_ref_list,
        ) = person
        if primary_name:
            primary_name = upgrade_name_16(self, primary_name)
        if alternate_names:
            alternate_names = upgrade_name_list_16(self, alternate_names)
        if address_list:
            address_list = upgrade_address_list_16(self, address_list)
        if media_list:
            media_list = upgrade_media_list_16(self, media_list)
        if attribute_list:
            attribute_list = upgrade_attribute_list_16(self, attribute_list)
        if lds_seal_list:
            lds_seal_list = upgrade_lds_seal_list_16(self, lds_seal_list)
        if source_list:
            new_citation_list = convert_source_list_to_citation_list_16(
                self, source_list
            )
        else:
            new_citation_list = []
        if person_ref_list:
            person_ref_list = upgrade_person_ref_list_16(self, person_ref_list)
        if event_ref_list:
            event_ref_list = upgrade_event_ref_list_16(self, event_ref_list)
        if (
            primary_name
            or alternate_names
            or address_list
            or media_list
            or attribute_list
            or lds_seal_list
            or source_list
            or person_ref_list
            or event_ref_list
        ):
            new_person = (
                handle,
                gramps_id,
                gender,
                primary_name,
                alternate_names,
                death_ref_index,
                birth_ref_index,
                event_ref_list,
                family_list,
                parent_family_list,
                media_list,
                address_list,
                attribute_list,
                urls,
                lds_seal_list,
                new_citation_list,
                note_list,
                change,
                tag_list,
                private,
                person_ref_list,
            )
            LOG.debug("      upgrade new_person %s" % [new_person])
            self._commit_raw(new_person, PERSON_KEY)
        self.update()

    LOG.debug(
        "%d persons upgraded with %d citations in %d seconds. "
        % (
            self.get_number_of_people(),
            self.cmap_index - start_num_citations,
            time.time() - start_time,
        )
    )
    data_upgradeobject[key2data[PERSON_KEY]] = (
        self.get_number_of_people(),
        self.cmap_index - start_num_citations,
        time.time() - start_time,
    )

    # ---------------------------------
    # Modify Media
    # ---------------------------------
    start_num_citations = self.cmap_index
    start_time = time.time()
    for media_handle in self.get_media_handles():
        media = self.get_raw_media_data(media_handle)
        LOG.debug("upgrade media object %s" % media[4])
        (
            handle,
            gramps_id,
            path,
            mime,
            desc,
            attribute_list,
            source_list,
            note_list,
            change,
            date,
            tag_list,
            private,
        ) = media
        new_citation_list = convert_source_list_to_citation_list_16(self, source_list)
        new_attribute_list = upgrade_attribute_list_16(self, attribute_list)

        new_media = (
            handle,
            gramps_id,
            path,
            mime,
            desc,
            new_attribute_list,
            new_citation_list,
            note_list,
            change,
            date,
            tag_list,
            private,
        )
        LOG.debug("      upgrade new_media %s" % [new_media])
        self._commit_raw(new_media, MEDIA_KEY)
        self.update()

    LOG.debug(
        "%d media objects upgraded with %d citations in %d seconds"
        % (
            self.get_number_of_media(),
            self.cmap_index - start_num_citations,
            int(time.time() - start_time),
        )
    )
    data_upgradeobject[key2data[MEDIA_KEY]] = (
        self.get_number_of_media(),
        self.cmap_index - start_num_citations,
        time.time() - start_time,
    )

    # ---------------------------------
    # Modify Places
    # ---------------------------------
    start_num_citations = self.cmap_index
    start_time = time.time()
    for place_handle in self.get_place_handles():
        place = self.get_raw_place_data(place_handle)
        LOG.debug("upgrade place %s" % place[2])
        (
            handle,
            gramps_id,
            title,
            longi,
            lat,
            main_loc,
            alt_loc,
            urls,
            media_list,
            source_list,
            note_list,
            change,
            private,
        ) = place
        if source_list:
            new_citation_list = convert_source_list_to_citation_list_16(
                self, source_list
            )
        else:
            new_citation_list = []
        if media_list:
            media_list = upgrade_media_list_16(self, media_list)
        if source_list or media_list:
            new_place = (
                handle,
                gramps_id,
                title,
                longi,
                lat,
                main_loc,
                alt_loc,
                urls,
                media_list,
                new_citation_list,
                note_list,
                change,
                private,
            )
            LOG.debug("      upgrade new_place %s" % [new_place])
            self._commit_raw(new_place, PLACE_KEY)
        self.update()

    LOG.debug(
        "%d places upgraded with %d citations in %d seconds. "
        % (
            self.get_number_of_places(),
            self.cmap_index - start_num_citations,
            time.time() - start_time,
        )
    )
    data_upgradeobject[key2data[PLACE_KEY]] = (
        self.get_number_of_places(),
        self.cmap_index - start_num_citations,
        time.time() - start_time,
    )

    # ---------------------------------
    # Modify Families
    # ---------------------------------
    start_num_citations = self.cmap_index
    start_time = time.time()
    for family_handle in self.get_family_handles():
        family = self.get_raw_family_data(family_handle)
        LOG.debug("upgrade family (gramps_id) %s" % family[1])
        (
            handle,
            gramps_id,
            father_handle,
            mother_handle,
            child_ref_list,
            the_type,
            event_ref_list,
            media_list,
            attribute_list,
            lds_seal_list,
            source_list,
            note_list,
            change,
            tag_list,
            private,
        ) = family
        if source_list:
            new_citation_list = convert_source_list_to_citation_list_16(
                self, source_list
            )
        else:
            new_citation_list = []
        if child_ref_list:
            child_ref_list = upgrade_child_ref_list_16(self, child_ref_list)
        if lds_seal_list:
            lds_seal_list = upgrade_lds_seal_list_16(self, lds_seal_list)
        if media_list:
            media_list = upgrade_media_list_16(self, media_list)
        if attribute_list:
            attribute_list = upgrade_attribute_list_16(self, attribute_list)
        if event_ref_list:
            event_ref_list = upgrade_event_ref_list_16(self, event_ref_list)
        if (
            source_list
            or media_list
            or child_ref_list
            or attribute_list
            or lds_seal_list
            or event_ref_list
        ):
            new_family = (
                handle,
                gramps_id,
                father_handle,
                mother_handle,
                child_ref_list,
                the_type,
                event_ref_list,
                media_list,
                attribute_list,
                lds_seal_list,
                new_citation_list,
                note_list,
                change,
                tag_list,
                private,
            )
            LOG.debug("      upgrade new_family %s" % [new_family])
            self._commit_raw(new_family, FAMILY_KEY)
        self.update()

    LOG.debug(
        "%d families upgraded with %d citations in %d seconds. "
        % (
            self.get_number_of_families(),
            self.cmap_index - start_num_citations,
            time.time() - start_time,
        )
    )
    data_upgradeobject[key2data[FAMILY_KEY]] = (
        self.get_number_of_families(),
        self.cmap_index - start_num_citations,
        time.time() - start_time,
    )
    # ---------------------------------
    # Modify Events
    # ---------------------------------
    start_num_citations = self.cmap_index
    start_time = time.time()
    for event_handle in self.get_event_handles():
        event = self.get_raw_event_data(event_handle)
        LOG.debug("upgrade event %s" % event[4])
        (
            handle,
            gramps_id,
            the_type,
            date,
            description,
            place,
            source_list,
            note_list,
            media_list,
            attribute_list,
            change,
            private,
        ) = event
        if source_list:
            new_citation_list = convert_source_list_to_citation_list_16(
                self, source_list
            )
        else:
            new_citation_list = []
        if attribute_list:
            attribute_list = upgrade_attribute_list_16(self, attribute_list)
        if media_list:
            media_list = upgrade_media_list_16(self, media_list)
        if source_list or attribute_list or media_list:
            new_event = (
                handle,
                gramps_id,
                the_type,
                date,
                description,
                place,
                new_citation_list,
                note_list,
                media_list,
                attribute_list,
                change,
                private,
            )
            LOG.debug("      upgrade new_event %s" % [new_event])
            self._commit_raw(new_event, EVENT_KEY)
        self.update()

    LOG.debug(
        "%d events upgraded with %d citations in %d seconds. "
        % (
            self.get_number_of_events(),
            self.cmap_index - start_num_citations,
            time.time() - start_time,
        )
    )
    data_upgradeobject[key2data[EVENT_KEY]] = (
        self.get_number_of_events(),
        self.cmap_index - start_num_citations,
        time.time() - start_time,
    )

    # ---------------------------------
    # Modify Repositories
    # ---------------------------------
    start_num_citations = self.cmap_index
    start_time = time.time()
    for repository_handle in self.get_repository_handles():
        repository = self.get_raw_repository_data(repository_handle)
        LOG.debug("upgrade repository %s" % repository[3])
        (
            handle,
            gramps_id,
            the_type,
            name,
            note_list,
            address_list,
            urls,
            change,
            private,
        ) = repository
        if address_list:
            address_list = upgrade_address_list_16(self, address_list)
        if address_list:
            new_repository = (
                handle,
                gramps_id,
                the_type,
                name,
                note_list,
                address_list,
                urls,
                change,
                private,
            )
            LOG.debug("      upgrade new_repository %s" % [new_repository])
            self._commit_raw(new_repository, REPOSITORY_KEY)
        self.update()

    LOG.debug(
        "%d repositories upgraded with %d citations in %d seconds. "
        % (
            self.get_number_of_repositories(),
            self.cmap_index - start_num_citations,
            time.time() - start_time,
        )
    )
    data_upgradeobject[key2data[REPOSITORY_KEY]] = (
        self.get_number_of_repositories(),
        self.cmap_index - start_num_citations,
        time.time() - start_time,
    )

    # ---------------------------------
    # Modify Source
    # ---------------------------------
    start_num_citations = self.cmap_index
    start_time = time.time()
    for source_handle in self.get_source_handles():
        source = self.get_raw_source_data(source_handle)
        LOG.debug("upgrade source %s" % source[2])
        (
            handle,
            gramps_id,
            title,
            author,
            pubinfo,
            note_list,
            media_list,
            abbrev,
            change,
            datamap,
            reporef_list,
            private,
        ) = source
        if media_list:
            media_list = upgrade_media_list_16(self, media_list)

        new_source = (
            handle,
            gramps_id,
            title,
            author,
            pubinfo,
            note_list,
            media_list,
            abbrev,
            change,
            datamap,
            reporef_list,
            private,
        )
        LOG.debug("      upgrade new_source %s" % [new_source])
        self._commit_raw(new_source, SOURCE_KEY)
        self.update()

    LOG.debug(
        "%d sources upgraded with %d citations in %d seconds"
        % (
            self.get_number_of_sources(),
            self.cmap_index - start_num_citations,
            int(time.time() - start_time),
        )
    )
    data_upgradeobject[key2data[SOURCE_KEY]] = (
        self.get_number_of_sources(),
        self.cmap_index - start_num_citations,
        time.time() - start_time,
    )

    # ---------------------------------
    # Example database from repository took:
    # 3403 events upgraded with 8 citations in 23 seconds. Backlinks took 1071 secs
    # actually 4 of these citations were from:
    # Media upgrade 4 citations upgraded in 4 seconds
    # by only doing the backlinks when there might be something to do,
    # improved to:
    # 3403 events upgraded with 8 citations in 19 seconds. Backlinks took 1348 secs
    # further improved by skipping debug logging:
    # 3403 events upgraded with 8 citations in 2 seconds. Backlinks took 167 secs

    # Number of new objects upgraded:
    #  2090  People        upgraded with   2092 citations in 2148 secs
    #   734  Families      upgraded with    735 citations in 768 secs
    #  3403  Events        upgraded with      4 citations in 212 secs
    #     7  Media Objects upgraded with      4 citations in 3 secs
    #   852  Places        upgraded with      0 citations in 39 secs

    # with reduced diagnostics
    # Number of new objects upgraded:
    #    73  People        upgraded with     76 citations in     74 secs
    #    35  Families      upgraded with     36 citations in     31 secs
    #  3403  Events        upgraded with      4 citations in      7 secs
    #     7  Media Objects upgraded with      4 citations in      3 secs
    #   852  Places        upgraded with      0 citations in      1 secs

    # without doing any backlinks
    # Number of new objects upgraded:
    #    73  People        upgraded with     76 citations in     43 secs
    #    35  Families      upgraded with     36 citations in     24 secs
    #  3403  Events        upgraded with      4 citations in      6 secs
    #     7  Media Objects upgraded with      4 citations in      2 secs
    #   852  Places        upgraded with      0 citations in      1 secs

    # another run about the same code:
    # Number of new objects upgraded:
    #    73  People        upgraded with     76 citations in     48 secs
    #    35  Families      upgraded with     36 citations in     21 secs
    #  3403  Events        upgraded with      4 citations in      9 secs
    #     7  Media Objects upgraded with      4 citations in      4 secs
    #   852  Places        upgraded with      0 citations in      1 secs

    # another run
    # Number of new objects upgraded:
    #    73  People        upgraded with     76 citations in     36 secs
    #    35  Families      upgraded with     36 citations in     18 secs
    #  3403  Events        upgraded with      4 citations in      9 secs
    #     7  Media Objects upgraded with      4 citations in      2 secs
    #   852  Places        upgraded with      0 citations in      1 secs

    # without incorrect nested tranaction structure:
    # Number of new objects upgraded:
    #    73  People        upgraded with     76 citations in      0 secs
    #    35  Families      upgraded with     36 citations in      0 secs
    #  3403  Events        upgraded with      4 citations in      0 secs
    #     7  Media Objects upgraded with      4 citations in      0 secs
    #   852  Places        upgraded with      0 citations in      0 secs

    self._txn_commit()
    # Bump up database version. Separate transaction to save metadata.
    self._set_metadata("version", 16)

    LOG.debug([data_upgradeobject])
    txt = _("Number of new objects upgraded:\n")
    for key in keyorder:
        try:
            txt += key2string[key] % {
                "n1": data_upgradeobject[key2data[key]][0],
                "n2": data_upgradeobject[key2data[key]][1],
                "n3": data_upgradeobject[key2data[key]][2],
            }
        except:
            txt += key2string[key]
    txt += _(
        "\n\nYou may want to run\n"
        "Tools -> Family Tree Processing -> Merge\n"
        "in order to merge citations that contain similar\n"
        "information"
    )
    InfoDialog(_("Upgrade Statistics"), txt, monospaced=True)  # TODO no-parent


def upgrade_media_list_16(self, media_list):
    new_media_list = []
    for media in media_list:
        (privacy, source_list, note_list, attribute_list, ref, rect) = media
        new_citation_list = convert_source_list_to_citation_list_16(self, source_list)
        new_attribute_list = upgrade_attribute_list_16(self, attribute_list)
        new_media = (
            privacy,
            new_citation_list,
            note_list,
            new_attribute_list,
            ref,
            rect,
        )
        new_media_list.append((new_media))
    return new_media_list


def upgrade_attribute_list_16(self, attribute_list):
    new_attribute_list = []
    for attribute in attribute_list:
        (privacy, source_list, note_list, the_type, value) = attribute
        new_citation_list = convert_source_list_to_citation_list_16(self, source_list)
        new_attribute = (privacy, new_citation_list, note_list, the_type, value)
        new_attribute_list.append((new_attribute))
    return new_attribute_list


def upgrade_child_ref_list_16(self, child_ref_list):
    new_child_ref_list = []
    for child_ref in child_ref_list:
        (privacy, source_list, note_list, ref, frel, mrel) = child_ref
        new_citation_list = convert_source_list_to_citation_list_16(self, source_list)
        new_child_ref = (privacy, new_citation_list, note_list, ref, frel, mrel)
        new_child_ref_list.append((new_child_ref))
    return new_child_ref_list


def upgrade_lds_seal_list_16(self, lds_seal_list):
    new_lds_seal_list = []
    for lds_seal in lds_seal_list:
        (
            source_list,
            note_list,
            date,
            type_,
            place,
            famc,
            temple,
            status,
            private,
        ) = lds_seal
        new_citation_list = convert_source_list_to_citation_list_16(self, source_list)
        new_lds_seal = (
            new_citation_list,
            note_list,
            date,
            type_,
            place,
            famc,
            temple,
            status,
            private,
        )
        new_lds_seal_list.append((new_lds_seal))
    return new_lds_seal_list


def upgrade_address_list_16(self, address_list):
    new_address_list = []
    for address in address_list:
        (privacy, source_list, note_list, date, location) = address
        new_citation_list = convert_source_list_to_citation_list_16(self, source_list)
        new_address = (privacy, new_citation_list, note_list, date, location)
        new_address_list.append((new_address))
    return new_address_list


def upgrade_name_list_16(self, name_list):
    new_name_list = []
    for name in name_list:
        new_name = upgrade_name_16(self, name)
        new_name_list.append((new_name))
    return new_name_list


def upgrade_name_16(self, name):
    (
        privacy,
        source_list,
        note,
        date,
        first_name,
        surname_list,
        suffix,
        title,
        name_type,
        group_as,
        sort_as,
        display_as,
        call,
        nick,
        famnick,
    ) = name
    new_citation_list = convert_source_list_to_citation_list_16(self, source_list)
    new_name = (
        privacy,
        new_citation_list,
        note,
        date,
        first_name,
        surname_list,
        suffix,
        title,
        name_type,
        group_as,
        sort_as,
        display_as,
        call,
        nick,
        famnick,
    )
    return new_name


def upgrade_person_ref_list_16(self, person_ref_list):
    new_person_ref_list = []
    for person_ref in person_ref_list:
        (privacy, source_list, note_list, ref, rel) = person_ref
        new_citation_list = convert_source_list_to_citation_list_16(self, source_list)
        new_person_ref = (privacy, new_citation_list, note_list, ref, rel)
        new_person_ref_list.append((new_person_ref))
    return new_person_ref_list


def upgrade_event_ref_list_16(self, event_ref_list):
    new_event_ref_list = []
    for event_ref in event_ref_list:
        (privacy, note_list, attribute_list, ref, role) = event_ref
        new_attribute_list = upgrade_attribute_list_16(self, attribute_list)
        new_event_ref = (privacy, note_list, new_attribute_list, ref, role)
        new_event_ref_list.append((new_event_ref))
    return new_event_ref_list


def convert_source_list_to_citation_list_16(self, source_list):
    citation_list = []
    for source in source_list:
        (date, private, note_list, confidence, ref, page) = source
        new_handle = create_id()
        new_media_list = []
        new_data_map = {}
        new_change = time.time()
        new_gramps_id = self.citation_prefix % self.cmap_index
        new_citation = (
            new_handle,
            new_gramps_id,
            date,
            page,
            confidence,
            ref,
            note_list,
            new_media_list,
            new_data_map,
            new_change,
            private,
        )
        citation_list.append((new_handle))
        self._commit_raw(new_citation, CITATION_KEY)
        self.cmap_index += 1
    return citation_list


def gramps_upgrade_15(self):
    """
    Upgrade database from version 14 to 15. This upgrade adds:

        * tagging
        * surname list
        * remove marker
    """
    self.set_serializer("blob")

    length = (
        self.get_number_of_notes()
        + self.get_number_of_people()
        + self.get_number_of_events()
        + self.get_number_of_families()
        + self.get_number_of_repositories()
        + self.get_number_of_media()
        + self.get_number_of_places()
        + self.get_number_of_sources()
    ) + 10
    self.set_total(length)
    self._txn_begin()
    self.tags = {}

    # ---------------------------------
    # Modify Person
    # ---------------------------------
    # Replace the old marker field with the new tag list field.
    for handle in self.get_person_handles():
        person = self.get_raw_person_data(handle)

        (
            junk_handle,  # 0
            gramps_id,  # 1
            gender,  # 2
            primary_name,  # 3
            alternate_names,  # 4
            death_ref_index,  # 5
            birth_ref_index,  # 6
            event_ref_list,  # 7
            family_list,  # 8
            parent_family_list,  # 9
            media_list,  # 10
            address_list,  # 11
            attribute_list,  # 12
            urls,  # 13
            ord_list,  # 14
            psource_list,  # 15
            pnote_list,  # 16
            change,  # 17
            marker,  # 18
            pprivate,  # 19
            person_ref_list,  # 20
        ) = person

        tag_handle = convert_marker(self, marker)
        if tag_handle:
            tags = [tag_handle]
        else:
            tags = []
        address_list = list(map(convert_address, address_list))
        new_primary_name = convert_name_15(primary_name)
        new_alternate_names = list(map(convert_name_15, alternate_names))
        new_person = (
            junk_handle,  # 0
            gramps_id,  # 1
            gender,  # 2
            new_primary_name,  # 3
            new_alternate_names,  # 4
            death_ref_index,  # 5
            birth_ref_index,  # 6
            event_ref_list,  # 7
            family_list,  # 8
            parent_family_list,  # 9
            media_list,  # 10
            address_list,  # 11
            attribute_list,  # 12
            urls,  # 13
            ord_list,  # 14
            psource_list,  # 15
            pnote_list,  # 16
            change,  # 17
            tags,  # 18
            pprivate,  # 19
            person_ref_list,  # 20
        )

        self._commit_raw(new_person, PERSON_KEY)
        self.update()

    # ---------------------------------
    # Modify Family
    # ---------------------------------
    # Replace the old marker field with the new tag list field.
    for handle in self.get_family_handles():
        family = self.get_raw_family_data(handle)
        new_family = list(family)
        tag_handle = convert_marker(self, new_family[13])
        if tag_handle:
            new_family[13] = [tag_handle]
        else:
            new_family[13] = []
        new_family = tuple(new_family)
        self._commit_raw(new_family, FAMILY_KEY)
        self.update()

    # ---------------------------------
    # Modify Note
    # ---------------------------------
    # Replace the old marker field with the new tag list field.
    for handle in self.get_note_handles():
        note = self.get_raw_note_data(handle)
        new_note = list(note)
        tag_handle = convert_marker(self, new_note[6])
        if tag_handle:
            new_note[6] = [tag_handle]
        else:
            new_note[6] = []
        new_note = tuple(new_note)
        self._commit_raw(new_note, NOTE_KEY)
        self.update()

    # ---------------------------------
    # Modify Media object
    # ---------------------------------
    # Replace the old marker field with the new tag list field.
    for handle in self.get_media_handles():
        media = self.get_raw_media_data(handle)
        new_media = list(media)
        new_media[10] = []
        new_media = tuple(new_media)
        self._commit_raw(new_media, MEDIA_KEY)
        self.update()

    # ---------------------------------
    # Modify Event
    # ---------------------------------
    # Replace the old marker field with the new tag list field.
    for handle in self.get_event_handles():
        event = self.get_raw_event_data(handle)
        new_event = list(event)
        new_event = new_event[:11] + new_event[12:]
        # new_event[11] = []
        new_event = tuple(new_event)
        self._commit_raw(new_event, EVENT_KEY)
        self.update()

    # ---------------------------------
    # Modify Place
    # ---------------------------------
    # Remove the old marker field, set new locality.
    for handle in self.get_place_handles():
        place = self.get_raw_place_data(handle)
        new_place = list(place)
        if new_place[5] is not None:
            new_place[5] = convert_location(new_place[5])
        new_place[6] = list(map(convert_location, new_place[6]))
        new_place = new_place[:12] + new_place[13:]
        new_place = tuple(new_place)
        self._commit_raw(new_place, PLACE_KEY)
        self.update()

    # ---------------------------------
    # Modify Source
    # ---------------------------------
    # Remove the old marker field.
    for handle in self.get_source_handles():
        source = self.get_raw_source_data(handle)
        new_source = list(source)
        new_source = new_source[:11] + new_source[12:]
        new_source = tuple(new_source)
        self._commit_raw(new_source, SOURCE_KEY)
        self.update()

    # ---------------------------------
    # Modify Repository
    # ---------------------------------
    # Remove the old marker field, set new locality.
    for handle in self.get_repository_handles():
        repository = self.get_raw_repository_data(handle)
        new_repository = list(repository)
        new_repository = new_repository[:8] + new_repository[9:]
        new_repository[5] = list(map(convert_address, new_repository[5]))
        new_repository = tuple(new_repository)
        self._commit_raw(new_repository, REPOSITORY_KEY)
        self.update()

    self._txn_commit()
    # Bump up database version. Separate transaction to save metadata.
    self._set_metadata("version", 15)


def convert_marker(self, marker_field):
    """Convert a marker into a tag."""
    marker = MarkerType()
    marker.unserialize(marker_field)
    tag_name = str(marker)

    if tag_name != "":
        if tag_name not in self.tags:
            tag = Tag()
            handle = create_id()
            tag.set_handle(handle)
            tag.set_change_time(time.time())
            tag.set_name(tag_name)
            tag.set_priority(len(self.tags))
            self._commit_raw(tag.serialize(), TAG_KEY)
            self.tags[tag_name] = handle
        return self.tags[tag_name]
    else:
        return None


def convert_locbase(loc):
    """Convert location base to include an empty locality field."""
    return tuple([loc[0], ""] + list(loc[1:]))


def convert_location(loc):
    """Convert a location into the new format."""
    return (convert_locbase(loc[0]), loc[1])


def convert_address(addr):
    """Convert an address into the new format."""
    return (addr[0], addr[1], addr[2], addr[3], convert_locbase(addr[4]))


def convert_name_15(name):
    (
        privacy,
        source_list,
        note_list,
        date,
        first_name,
        surname,
        suffix,
        title,
        name_type,
        prefix,
        patronymic,
        group_as,
        sort_as,
        display_as,
        call,
    ) = name

    connector = ""
    origintype = (NameOriginType.NONE, "")
    patorigintype = (NameOriginType.PATRONYMIC, "")

    if patronymic.strip() == "":
        # no patronymic, create a single surname
        surname_list = [(surname, prefix, True, origintype, connector)]
    else:
        # a patronymic, if no surname or equal as patronymic, a single surname
        if (surname.strip() == "") or (surname == patronymic and prefix == ""):
            surname_list = [(patronymic, prefix, True, patorigintype, connector)]
        else:
            # two surnames, first patronymic, then surname which is primary
            surname_list = [
                (patronymic, "", False, patorigintype, ""),
                (surname, prefix, True, origintype, connector),
            ]

    # return new value, add two empty strings for nick and family nick
    return (
        privacy,
        source_list,
        note_list,
        date,
        first_name,
        surname_list,
        suffix,
        title,
        name_type,
        group_as,
        sort_as,
        display_as,
        call,
        "",
        "",
    )


def gramps_upgrade_14(self):
    """Upgrade database from version 13 to 14."""
    # This upgrade modifies notes and dates
    self.set_serializer("blob")

    length = (
        self.get_number_of_notes()
        + self.get_number_of_people()
        + self.get_number_of_events()
        + self.get_number_of_families()
        + self.get_number_of_repositories()
        + self.get_number_of_media()
        + self.get_number_of_places()
        + self.get_number_of_sources()
    )
    self.set_total(length)
    self._txn_begin()

    # ---------------------------------
    # Modify Notes
    # ---------------------------------
    # replace clear text with StyledText in Notes
    for handle in self.get_note_handles():
        note = self.get_raw_note_data(handle)
        (
            junk_handle,
            gramps_id,
            text,
            format_,
            note_type,
            change,
            marker,
            private,
        ) = note
        styled_text = (text, [])
        new_note = (
            handle,
            gramps_id,
            styled_text,
            format_,
            note_type,
            change,
            marker,
            private,
        )
        self._commit_raw(new_note, NOTE_KEY)
        self.update()

    # ---------------------------------
    # Modify Event
    # ---------------------------------
    # update dates with newyear
    for handle in self.get_event_handles():
        event = self.get_raw_event_data(handle)
        (
            junk_handle,
            gramps_id,
            the_type,
            date,
            description,
            place,
            source_list,
            note_list,
            media_list,
            attribute_list,
            change,
            marker,
            private,
        ) = event
        new_date = convert_date_14(date)
        new_source_list = new_source_list_14(source_list)
        new_media_list = new_media_list_14(media_list)
        new_attribute_list = new_attribute_list_14(attribute_list)
        new_event = (
            junk_handle,
            gramps_id,
            the_type,
            new_date,
            description,
            place,
            new_source_list,
            note_list,
            new_media_list,
            new_attribute_list,
            change,
            marker,
            private,
        )
        self._commit_raw(new_event, EVENT_KEY)
        self.update()

    # ---------------------------------
    # Modify Person
    # ---------------------------------
    # update dates with newyear
    for handle in self.get_person_handles():
        person = self.get_raw_person_data(handle)
        (
            junk_handle,  # 0
            gramps_id,  # 1
            gender,  # 2
            primary_name,  # 3
            alternate_names,  # 4
            death_ref_index,  # 5
            birth_ref_index,  # 6
            event_ref_list,  # 7
            family_list,  # 8
            parent_family_list,  # 9
            media_list,  # 10
            address_list,  # 11
            attribute_list,  # 12
            urls,  # 13
            lds_ord_list,  # 14
            psource_list,  # 15
            pnote_list,  # 16
            change,  # 17
            marker,  # 18
            pprivate,  # 19
            person_ref_list,  # 20
        ) = person

        new_address_list = []
        for address in address_list:
            (privacy, asource_list, anote_list, date, location) = address
            new_date = convert_date_14(date)
            new_asource_list = new_source_list_14(asource_list)
            new_address_list.append(
                (privacy, new_asource_list, anote_list, new_date, location)
            )
        new_ord_list = []
        for ldsord in lds_ord_list:
            (
                lsource_list,
                lnote_list,
                date,
                type_,
                place,
                famc,
                temple,
                status,
                lprivate,
            ) = ldsord
            new_date = convert_date_14(date)
            new_lsource_list = new_source_list_14(lsource_list)
            new_ord_list.append(
                (
                    new_lsource_list,
                    lnote_list,
                    new_date,
                    type_,
                    place,
                    famc,
                    temple,
                    status,
                    lprivate,
                )
            )

        new_primary_name = convert_name_14(primary_name)

        new_alternate_names = [convert_name_14(name) for name in alternate_names]

        new_media_list = new_media_list_14(media_list)
        new_psource_list = new_source_list_14(psource_list)
        new_attribute_list = new_attribute_list_14(attribute_list)
        new_person_ref_list = new_person_ref_list_14(person_ref_list)

        new_person = (
            junk_handle,  # 0
            gramps_id,  # 1
            gender,  # 2
            new_primary_name,  # 3
            new_alternate_names,  # 4
            death_ref_index,  # 5
            birth_ref_index,  # 6
            event_ref_list,  # 7
            family_list,  # 8
            parent_family_list,  # 9
            new_media_list,  # 10
            new_address_list,  # 11
            new_attribute_list,  # 12
            urls,  # 13
            new_ord_list,  # 14
            new_psource_list,  # 15
            pnote_list,  # 16
            change,  # 17
            marker,  # 18
            pprivate,  # 19
            new_person_ref_list,  # 20
        )

        self._commit_raw(new_person, PERSON_KEY)
        self.update()

    # ---------------------------------
    # Modify Family
    # ---------------------------------
    # update dates with newyear
    for handle in self.get_family_handles():
        family = self.get_raw_family_data(handle)
        (
            junk_handle,
            gramps_id,
            father_handle,
            mother_handle,
            child_ref_list,
            the_type,
            event_ref_list,
            media_list,
            attribute_list,
            lds_seal_list,
            source_list,
            note_list,
            change,
            marker,
            private,
        ) = family
        new_child_ref_list = new_child_ref_list_14(child_ref_list)
        new_media_list = new_media_list_14(media_list)
        new_source_list = new_source_list_14(source_list)
        new_attribute_list = new_attribute_list_14(attribute_list)
        new_seal_list = []
        for ldsord in lds_seal_list:
            (
                lsource_list,
                lnote_list,
                date,
                type_,
                place,
                famc,
                temple,
                status,
                lprivate,
            ) = ldsord
            new_date = convert_date_14(date)
            new_lsource_list = new_source_list_14(lsource_list)
            new_seal_list.append(
                (
                    new_lsource_list,
                    lnote_list,
                    new_date,
                    type_,
                    place,
                    famc,
                    temple,
                    status,
                    lprivate,
                )
            )

        new_family = (
            junk_handle,
            gramps_id,
            father_handle,
            mother_handle,
            new_child_ref_list,
            the_type,
            event_ref_list,
            new_media_list,
            new_attribute_list,
            new_seal_list,
            new_source_list,
            note_list,
            change,
            marker,
            private,
        )

        self._commit_raw(new_family, FAMILY_KEY)
        self.update()

    # ---------------------------------
    # Modify Repository
    # ---------------------------------
    # update dates with newyear
    for handle in self.get_repository_handles():
        repository = self.get_raw_repository_data(handle)
        # address
        (
            junk_handle,
            gramps_id,
            the_type,
            name,
            note_list,
            address_list,
            urls,
            change,
            marker,
            private,
        ) = repository

        new_address_list = []
        for address in address_list:
            (privacy, asource_list, anote_list, date, location) = address
            new_date = convert_date_14(date)
            new_asource_list = new_source_list_14(asource_list)
            new_address_list.append(
                (privacy, new_asource_list, anote_list, new_date, location)
            )

        new_repository = (
            junk_handle,
            gramps_id,
            the_type,
            name,
            note_list,
            new_address_list,
            urls,
            change,
            marker,
            private,
        )

        self._commit_raw(new_repository, REPOSITORY_KEY)
        self.update()

    # ---------------------------------
    # Modify Media
    # ---------------------------------
    for media_handle in self.get_media_handles():
        media = self.get_raw_media_data(media_handle)
        (
            handle,
            gramps_id,
            path,
            mime,
            desc,
            attribute_list,
            source_list,
            note_list,
            change,
            date,
            marker,
            private,
        ) = media
        new_source_list = new_source_list_14(source_list)
        new_date = convert_date_14(date)
        new_media = (
            handle,
            gramps_id,
            path,
            mime,
            desc,
            attribute_list,
            new_source_list,
            note_list,
            change,
            new_date,
            marker,
            private,
        )

        self._commit_raw(new_media, MEDIA_KEY)
        self.update()

    # ---------------------------------
    # Modify Place
    # ---------------------------------
    for place_handle in self.get_place_handles():
        place = self.get_raw_place_data(place_handle)
        (
            handle,
            gramps_id,
            title,
            longi,
            lat,
            main_loc,
            alt_loc,
            urls,
            media_list,
            source_list,
            note_list,
            change,
            marker,
            private,
        ) = place
        new_media_list = new_media_list_14(media_list)
        new_source_list = new_source_list_14(source_list)
        new_place = (
            handle,
            gramps_id,
            title,
            longi,
            lat,
            main_loc,
            alt_loc,
            urls,
            new_media_list,
            new_source_list,
            note_list,
            change,
            marker,
            private,
        )

        self._commit_raw(new_place, PLACE_KEY)
        self.update()

    # ---------------------------------
    # Modify Source
    # ---------------------------------
    for source_handle in self.get_source_handles():
        source = self.get_raw_source_data(source_handle)
        (
            handle,
            gramps_id,
            title,
            author,
            pubinfo,
            note_list,
            media_list,
            abbrev,
            change,
            datamap,
            reporef_list,
            marker,
            private,
        ) = source
        new_media_list = new_media_list_14(media_list)
        new_source = (
            handle,
            gramps_id,
            title,
            author,
            pubinfo,
            note_list,
            new_media_list,
            abbrev,
            change,
            datamap,
            reporef_list,
            marker,
            private,
        )

        self._commit_raw(new_source, SOURCE_KEY)
        self.update()

    self._txn_commit()
    # Bump up database version. Separate transaction to save metadata.
    self._set_metadata("version", 14)


def new_source_list_14(source_list):
    new_source_list = []
    for source in source_list:
        (date, private, note_list, confidence, ref, page) = source
        new_date = convert_date_14(date)
        new_source_list.append((new_date, private, note_list, confidence, ref, page))
    return new_source_list


def new_attribute_list_14(attribute_list):
    new_attribute_list = []
    for attribute in attribute_list:
        (private, asource_list, note_list, the_type, value) = attribute
        new_asource_list = new_source_list_14(asource_list)
        new_attribute_list.append(
            (private, new_asource_list, note_list, the_type, value)
        )
    return new_attribute_list


def new_media_list_14(media_list):
    # ---------------------------------
    # Event Media list
    # ---------------------------------
    new_media_list = []
    for media in media_list:
        (private, source_list, note_list, attribute_list, ref, role) = media
        new_source_list = new_source_list_14(source_list)
        new_attribute_list = new_attribute_list_14(attribute_list)
        new_media_list.append(
            (private, new_source_list, note_list, new_attribute_list, ref, role)
        )
    return new_media_list


def new_person_ref_list_14(person_ref_list):
    new_person_ref_list = []
    for person_ref in person_ref_list:
        (private, source_list, note_list, ref, rel) = person_ref
        new_source_list = new_source_list_14(source_list)
        new_person_ref_list.append((private, new_source_list, note_list, ref, rel))
    return new_person_ref_list


def new_child_ref_list_14(child_ref_list):
    new_child_ref_list = []
    for data in child_ref_list:
        (private, source_list, note_list, ref, frel, mrel) = data
        new_source_list = new_source_list_14(source_list)
        new_child_ref_list.append(
            (private, new_source_list, note_list, ref, frel, mrel)
        )
    return new_child_ref_list


def convert_date_14(date):
    if date:
        (calendar, modifier, quality, dateval, text, sortval) = date
        return (calendar, modifier, quality, dateval, text, sortval, 0)
    else:
        return None


def convert_name_14(name):
    (
        privacy,
        source_list,
        note_list,
        date,
        first_name,
        surname,
        suffix,
        title,
        name_type,
        prefix,
        patronymic,
        group_as,
        sort_as,
        display_as,
        call,
    ) = name
    new_date = convert_date_14(date)
    new_source_list = new_source_list_14(source_list)
    return (
        privacy,
        new_source_list,
        note_list,
        new_date,
        first_name,
        surname,
        suffix,
        title,
        name_type,
        prefix,
        patronymic,
        group_as,
        sort_as,
        display_as,
        call,
    )


def make_zip_backup(dirname):
    """
    This backs up the db files so an upgrade can be (manually) undone.
    """
    LOG.debug("Make backup prior to schema upgrade")
    import zipfile

    # In Windows reserved characters is "<>:"/\|?*"
    reserved_char = r':,<>"/\|?* '
    replace_char = "-__________"
    filepath = os.path.join(dirname, NAME_FILE)
    with open(filepath, "r", encoding="utf8") as name_file:
        title = name_file.readline().strip()
    trans = title.maketrans(reserved_char, replace_char)
    title = title.translate(trans)

    if not os.access(dirname, os.W_OK):
        LOG.warning("Can't write technical DB backup for %s", title)
        return
    (grampsdb_path, db_code) = os.path.split(dirname)
    dotgramps_path = os.path.dirname(grampsdb_path)
    zipname = title + time.strftime("_%Y-%m-%d_%H-%M-%S") + ".zip"
    zippath = os.path.join(dotgramps_path, zipname)
    with zipfile.ZipFile(zippath, "w") as myzip:
        for filename in os.listdir(dirname):
            pathname = os.path.join(dirname, filename)
            myzip.write(pathname, os.path.join(db_code, filename))
    LOG.warning(
        "If upgrade and loading the Family Tree works, you can "
        "delete the zip file at %s",
        zippath,
    )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import GLib

from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.utils.db import family_name, get_participant_from_event


# -------------------------------------------------------------------------
#
# BackRefModel
#
# -------------------------------------------------------------------------
class BackRefModel(Gtk.ListStore):
    dispstr = _("%(part1)s - %(part2)s")

    def __init__(self, sref_list, db):
        Gtk.ListStore.__init__(self, str, str, str, str, str)
        self.db = db
        self.sref_list = sref_list
        self.count = 0
        self.loading = False
        self.hndl_iter = {}
        self.idle = GLib.idle_add(self.load_model().__next__)

    def destroy(self):
        if self.loading:
            GLib.source_remove(self.idle)

    def load_model(self):
        """
        Objects can have very large backreferences. To avoid blocking the
        interface up to the moment that the model is created, this method is
        called via GLib.idle_add.
        WARNING: a consequence of above is that loading can still be happening
            while the GUI using this model is no longer used. Disconnect any
            methods before closing the GUI.
        """
        self.loading = True
        self.count = 0
        for ref in self.sref_list:
            self.count += 1
            dtype = ref[0]
            if dtype == "Person":
                p = self.db.get_person_from_handle(ref[1])
                if not p:
                    continue
                gid = p.gramps_id
                handle = p.handle
                name = name_displayer.display(p)
            elif dtype == "Family":
                p = self.db.get_family_from_handle(ref[1])
                if not p:
                    continue
                gid = p.gramps_id
                handle = p.handle
                name = family_name(p, self.db)
            elif dtype == "Source":
                p = self.db.get_source_from_handle(ref[1])
                if not p:
                    continue
                gid = p.gramps_id
                handle = p.handle
                name = p.get_title()
            elif dtype == "Citation":
                p = self.db.get_citation_from_handle(ref[1])
                if not p:
                    continue
                gid = p.gramps_id
                handle = p.handle
                name = p.get_page()
            elif dtype == "Event":
                p = self.db.get_event_from_handle(ref[1])
                if not p:
                    continue
                gid = p.gramps_id
                handle = p.handle
                name = p.get_description()
                if name:
                    name = self.dispstr % {"part1": str(p.get_type()), "part2": name}
                else:
                    name = str(p.get_type())
                part = get_participant_from_event(self.db, ref[1])
                if part:
                    name = self.dispstr % {"part1": name, "part2": part}
            elif dtype == "Place":
                p = self.db.get_place_from_handle(ref[1])
                if not p:
                    continue
                name = place_displayer.display(self.db, p)
                gid = p.gramps_id
                handle = p.handle
            elif dtype == "Repository":
                p = self.db.get_repository_from_handle(ref[1])
                if not p:
                    continue
                name = p.get_name()
                gid = p.gramps_id
                handle = p.handle
            elif dtype == "Note":
                p = self.db.get_note_from_handle(ref[1])
                if not p:
                    continue
                name = " ".join(p.get().split())
                if len(name) > 80:
                    name = name[:80] + "..."
                gid = p.gramps_id
                handle = p.handle
            else:
                p = self.db.get_media_from_handle(ref[1])
                if not p:
                    continue
                name = p.get_description()
                gid = p.gramps_id
                handle = p.handle

            # dtype is the class name, i.e. is English
            # We need to use localized string in the model.
            # we also need to keep class names to get the object type,
            # but we don't need to show that in the view.

            self.hndl_iter[ref[1]] = self.append(
                row=[_(dtype), gid, name, handle, dtype]
            )
            yield True
        self.loading = False
        yield False

    def delete_row(self, hndl_list):
        """Remove rows of the model based on the handles"""
        for hndl in hndl_list:
            miter = self.hndl_iter.get(hndl, None)
            if miter is not None:
                self.remove(miter)
                del self.hndl_iter[hndl]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2006-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Child Reference class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .childreftype import ChildRefType
from .citationbase import CitationBase
from .const import DIFFERENT, EQUAL, IDENTICAL
from .notebase import NoteBase
from .privacybase import PrivacyBase
from .refbase import RefBase
from .secondaryobj import SecondaryObject

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# ChildRef
#
# -------------------------------------------------------------------------
class ChildRef(SecondaryObject, PrivacyBase, CitationBase, NoteBase, RefBase):
    """
    A class for tracking information about how a child relates to their parents.
    """

    def __init__(self, source=None):
        PrivacyBase.__init__(self, source)
        CitationBase.__init__(self, source)
        NoteBase.__init__(self, source)
        RefBase.__init__(self, source)
        if source:
            self.frel = ChildRefType(source.frel)
            self.mrel = ChildRefType(source.mrel)
        else:
            self.frel = ChildRefType()
            self.mrel = ChildRefType()

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            PrivacyBase.serialize(self),
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            RefBase.serialize(self),
            self.frel.serialize(),
            self.mrel.serialize(),
        )

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (privacy, citation_list, note_list, ref, frel, mrel) = data
        PrivacyBase.unserialize(self, privacy)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        RefBase.unserialize(self, ref)
        self.frel = ChildRefType()
        self.frel.unserialize(frel)
        self.mrel = ChildRefType()
        self.mrel.unserialize(mrel)
        return self

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        return {
            "type": "object",
            "title": _("Child Reference"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "private": {"type": "boolean", "title": _("Private")},
                "citation_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Citations"),
                },
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "ref": {
                    "type": "string",
                    "maxLength": 50,
                    "title": _("Handle"),
                },
                "frel": ChildRefType.get_schema(),
                "mrel": ChildRefType.get_schema(),
            },
        }

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [str(self.frel), str(self.mrel)]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return []

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return []

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        ret = (
            self.get_referenced_note_handles() + self.get_referenced_citation_handles()
        )
        if self.ref:
            ret += [("Person", self.ref)]
        return ret

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through their
        children, reference primary objects..

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return []

    def is_equivalent(self, other):
        """
        Return if this child reference is equivalent, that is agrees in hlink,
        to other.

        :param other: The childref to compare this one to.
        :type other: ChildRef
        :returns: Constant indicating degree of equivalence.
        :rtype: int
        """
        if self.ref != other.ref:
            return DIFFERENT
        if self.is_equal(other):
            return IDENTICAL
        return EQUAL

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this child reference.

        Lost: hlink, mrel and frel of acquisition.

        :param acquisition: The childref to merge with the present childref.
        :type acquisition: ChildRef
        """
        self._merge_privacy(acquisition)
        self._merge_note_list(acquisition)
        self._merge_citation_list(acquisition)
        if (self.mrel != acquisition.mrel) or (self.frel != acquisition.frel):
            if self.mrel == ChildRefType.UNKNOWN:
                self.set_mother_relation(acquisition.mrel)
            if self.frel == ChildRefType.UNKNOWN:
                self.set_father_relation(acquisition.frel)

    def set_mother_relation(self, rel):
        """Set relation between the person and mother."""
        self.mrel.set(rel)

    def get_mother_relation(self):
        """Return the relation between the person and mother."""
        return self.mrel

    def set_father_relation(self, frel):
        """Set relation between the person and father."""
        self.frel.set(frel)

    def get_father_relation(self):
        """Return the relation between the person and father."""
        return self.frel



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2007  Donald N. Allingham
# Copyright (C) 2007-2012  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2012       Nick Hall
# Copyright (C) 2011-2016  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""GUI dialog for creating and managing books"""

# Written by Alex Roitman,
# largely based on the BaseDoc classes by Don Allingham

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".Book")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GObject

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ...listmodel import ListModel
from gramps.gen.errors import FilterError, ReportError
from gramps.gen.const import URL_MANUAL_PAGE
from ...display import display_help
from ...pluginmanager import GuiPluginManager
from ...dialog import WarningDialog, ErrorDialog, QuestionDialog2
from gramps.gen.plug.menu import PersonOption, FamilyOption
from gramps.gen.plug.docgen import StyleSheet
from ...managedwindow import ManagedWindow, set_titles
from ...glade import Glade
from ...utils import is_right_click, open_file_with_default_application
from ...user import User
from .. import make_gui_option

# Import from specific modules in ReportBase
from gramps.gen.plug.report import BookList, Book, BookItem, append_styles
from gramps.gen.plug.report import CATEGORY_BOOK, book_categories
from gramps.gen.plug.report._options import ReportOptions
from ._reportdialog import ReportDialog
from ._docreportdialog import DocReportDialog

# ------------------------------------------------------------------------
#
# Private Constants
#
# ------------------------------------------------------------------------
_UNSUPPORTED = _("Unsupported")

_RETURN = Gdk.keyval_from_name("Return")
_KP_ENTER = Gdk.keyval_from_name("KP_Enter")
WIKI_HELP_PAGE = URL_MANUAL_PAGE + "_-_Reports_-_part_3"
WIKI_HELP_SEC = _("Books")
GENERATE_WIKI_HELP_SEC = _("Generate_Book_dialog")


# ------------------------------------------------------------------------
#
# Private Functions
#
# ------------------------------------------------------------------------
def _initialize_options(options, dbstate, uistate):
    """
    Validates all options by making sure that their values are consistent with
    the database.

    menu: The Menu class
    dbase: the database the options will be applied to
    """
    if not hasattr(options, "menu"):
        return
    dbase = dbstate.get_database()
    if dbase.get_total() == 0:
        return
    menu = options.menu

    for name in menu.get_all_option_names():
        option = menu.get_option_by_name(name)
        value = option.get_value()

        if isinstance(option, PersonOption):
            if not dbase.get_person_from_gramps_id(value):
                person_handle = uistate.get_active("Person")
                person = dbase.get_person_from_handle(person_handle)
                option.set_value(person.get_gramps_id())

        elif isinstance(option, FamilyOption):
            if not dbase.get_family_from_gramps_id(value):
                person_handle = uistate.get_active("Person")
                person = dbase.get_person_from_handle(person_handle)
                if person is None:
                    continue
                family_list = person.get_family_handle_list()
                if family_list:
                    family_handle = family_list[0]
                else:
                    try:
                        family_handle = next(dbase.iter_family_handles())
                    except StopIteration:
                        family_handle = None
                if family_handle:
                    family = dbase.get_family_from_handle(family_handle)
                    option.set_value(family.get_gramps_id())
                else:
                    print("No family specified for ", name)


# ------------------------------------------------------------------------
#
# BookListDisplay class
#
# ------------------------------------------------------------------------
class BookListDisplay:
    """
    Interface into a dialog with the list of available books.

    Allows the user to select and/or delete a book from the list.
    """

    def __init__(self, booklist, nodelete=False, dosave=False, parent=None):
        """
        Create a BookListDisplay object that displays the books in BookList.

        booklist:   books that are displayed -- a :class:`.BookList` instance
        nodelete:   if True then the Delete button is hidden
        dosave:     if True then the book list is flagged to be saved if needed
        """

        self.booklist = booklist
        self.dosave = dosave
        self.xml = Glade("book.glade", toplevel="book")
        self.top = self.xml.toplevel
        self.unsaved_changes = False

        set_titles(self.top, self.xml.get_object("title2"), _("Available Books"))

        if nodelete:
            delete_button = self.xml.get_object("delete_button")
            delete_button.hide()
        self.xml.connect_signals(
            {
                "on_booklist_cancel_clicked": self.on_booklist_cancel_clicked,
                "on_booklist_ok_clicked": self.on_booklist_ok_clicked,
                "on_booklist_delete_clicked": self.on_booklist_delete_clicked,
                "on_book_ok_clicked": self.do_nothing,
                "on_book_help_clicked": self.do_nothing,
                "destroy_passed_object": self.do_nothing,
                "on_setup_clicked": self.do_nothing,
                "on_down_clicked": self.do_nothing,
                "on_up_clicked": self.do_nothing,
                "on_remove_clicked": self.do_nothing,
                "on_add_clicked": self.do_nothing,
                "on_edit_clicked": self.do_nothing,
                "on_open_clicked": self.do_nothing,
                "on_save_clicked": self.do_nothing,
                "on_clear_clicked": self.do_nothing,
            }
        )
        self.guilistbooks = self.xml.get_object("list")
        self.guilistbooks.connect("button-press-event", self.on_button_press)
        self.guilistbooks.connect("key-press-event", self.on_key_pressed)
        self.blist = ListModel(
            self.guilistbooks,
            [("Name", -1, 10)],
        )

        self.redraw()
        self.selection = None
        self.top.set_transient_for(parent)
        self.top.run()

    def redraw(self):
        """Redraws the list of currently available books"""

        self.blist.model.clear()
        names = self.booklist.get_book_names()
        if not len(names):
            return
        for name in names:
            the_iter = self.blist.add([name])
        if the_iter:
            self.blist.selection.select_iter(the_iter)

    def on_booklist_ok_clicked(self, obj):
        """
        Return selected book.
        Also marks the current list to be saved into the xml file, if needed.
        """
        store, the_iter = self.blist.get_selected()
        if the_iter:
            data = self.blist.get_data(the_iter, [0])
            self.selection = self.booklist.get_book(str(data[0]))
        if self.dosave and self.unsaved_changes:
            self.booklist.set_needs_saving(True)

    def on_booklist_delete_clicked(self, obj):
        """
        Deletes selected book from the list.

        This change is not final. OK button has to be clicked to save the list.
        """
        store, the_iter = self.blist.get_selected()
        if not the_iter:
            return
        data = self.blist.get_data(the_iter, [0])
        self.booklist.delete_book(str(data[0]))
        self.blist.remove(the_iter)
        self.unsaved_changes = True
        self.top.run()

    def on_booklist_cancel_clicked(self, *obj):
        """cancel the booklist dialog"""
        if self.unsaved_changes:
            qqq = QuestionDialog2(
                _("Discard Unsaved Changes"),
                _("You have made changes which have not been saved."),
                _("Proceed"),
                _("Cancel"),
                parent=self.top,
            )
            if not qqq.run():
                self.top.run()

    def on_button_press(self, obj, event):
        """
        Checks for a double click event. In the list, we want to
        treat a double click as if it was OK button press.
        """
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            store, the_iter = self.blist.get_selected()
            if not the_iter:
                return False
            self.on_booklist_ok_clicked(obj)
            # emit OK response on dialog to close it automatically
            self.top.response(-5)
            return True
        return False

    def on_key_pressed(self, obj, event):
        """
        Handles the return key being pressed on list. If the key is pressed,
        the Edit button handler is called
        """
        if event.type == Gdk.EventType.KEY_PRESS:
            if event.keyval in (_RETURN, _KP_ENTER):
                self.on_booklist_ok_clicked(obj)
                # emit OK response on dialog to close it automatically
                self.top.response(-5)
                return True
        return False

    def do_nothing(self, obj):
        """do nothing"""
        pass


# ------------------------------------------------------------------------
#
# Book Options
#
# ------------------------------------------------------------------------
class BookOptions(ReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        ReportOptions.__init__(self, name, dbase)

        # Options specific for this report
        self.options_dict = {
            "bookname": "",
        }
        # TODO since the CLI code for the "book" generates its own "help" now,
        # the GUI code would be faster if it didn't list all the possible books
        self.options_help = {
            "bookname": (
                "=name",
                _("Name of the book. MANDATORY"),
                BookList("books.xml", dbase).get_book_names(),
                False,
            ),
        }


# -------------------------------------------------------------------------
#
# Book creation dialog
#
# -------------------------------------------------------------------------
class BookSelector(ManagedWindow):
    """
    Interface into a dialog setting up the book.

    Allows the user to add/remove/reorder/setup items for the current book
    and to clear/load/save/edit whole books.
    """

    def __init__(self, dbstate, uistate):
        self._db = dbstate.db
        self.dbstate = dbstate
        self.uistate = uistate
        self.title = _("Manage Books")
        self.file = "books.xml"

        ManagedWindow.__init__(self, uistate, [], self.__class__)

        self.xml = Glade("book.glade", toplevel="top")
        window = self.xml.toplevel

        title_label = self.xml.get_object("title")
        self.set_window(window, title_label, self.title)
        self.setup_configs("interface.bookselector", 700, 600)
        self.show()
        self.xml.connect_signals(
            {
                "on_add_clicked": self.on_add_clicked,
                "on_remove_clicked": self.on_remove_clicked,
                "on_up_clicked": self.on_up_clicked,
                "on_down_clicked": self.on_down_clicked,
                "on_setup_clicked": self.on_setup_clicked,
                "on_clear_clicked": self.on_clear_clicked,
                "on_save_clicked": self.on_save_clicked,
                "on_open_clicked": self.on_open_clicked,
                "on_edit_clicked": self.on_edit_clicked,
                "on_book_help_clicked": lambda x: display_help(
                    WIKI_HELP_PAGE, WIKI_HELP_SEC
                ),
                "on_book_ok_clicked": self.on_book_ok_clicked,
                "destroy_passed_object": self.on_close_clicked,
                # Insert dummy handlers for second top level in the glade file
                "on_booklist_ok_clicked": lambda _: None,
                "on_booklist_delete_clicked": lambda _: None,
                "on_booklist_cancel_clicked": lambda _: None,
                "on_booklist_ok_clicked": lambda _: None,
                "on_booklist_ok_clicked": lambda _: None,
            }
        )

        self.avail_tree = self.xml.get_object("avail_tree")
        self.book_tree = self.xml.get_object("book_tree")
        self.avail_tree.connect("button-press-event", self.avail_button_press)
        self.book_tree.connect("button-press-event", self.book_button_press)

        self.name_entry = self.xml.get_object("name_entry")
        self.name_entry.set_text(_("New Book"))

        avail_label = self.xml.get_object("avail_label")
        avail_label.set_text("<b>%s</b>" % _("_Available items"))
        avail_label.set_use_markup(True)
        avail_label.set_use_underline(True)
        book_label = self.xml.get_object("book_label")
        book_label.set_text("<b>%s</b>" % _("Current _book"))
        book_label.set_use_underline(True)
        book_label.set_use_markup(True)

        avail_titles = [(_("Name"), 0, 230), (_("Type"), 1, 80), ("", -1, 0)]

        book_titles = [
            (_("Item name"), -1, 230),
            (_("Type"), -1, 80),
            ("", -1, 0),
            (_("Subject"), -1, 50),
        ]

        self.avail_nr_cols = len(avail_titles)
        self.book_nr_cols = len(book_titles)

        self.avail_model = ListModel(self.avail_tree, avail_titles)
        self.book_model = ListModel(self.book_tree, book_titles)
        self.draw_avail_list()

        self.book = Book()
        self.book_list = BookList(self.file, self._db)
        self.book_list.set_needs_saving(False)  # just read in: no need to save

    def build_menu_names(self, obj):
        return (_("Book selection list"), self.title)

    def draw_avail_list(self):
        """
        Draw the list with the selections available for the book.

        The selections are read from the book item registry.
        """
        pmgr = GuiPluginManager.get_instance()
        regbi = pmgr.get_reg_bookitems()
        if not regbi:
            return

        available_reports = []
        for pdata in regbi:
            category = _UNSUPPORTED
            if pdata.supported and pdata.category in book_categories:
                category = book_categories[pdata.category]
            available_reports.append([pdata.name, category, pdata.id])
        for data in sorted(available_reports):
            new_iter = self.avail_model.add(data)

        self.avail_model.connect_model()

        if new_iter:
            self.avail_model.selection.select_iter(new_iter)
            path = self.avail_model.model.get_path(new_iter)
            col = self.avail_tree.get_column(0)
            self.avail_tree.scroll_to_cell(path, col, 1, 1, 0.0)

    def open_book(self, book):
        """
        Open the book: set the current set of selections to this book's items.

        book:   the book object to load.
        """
        if book.get_paper_name():
            self.book.set_paper_name(book.get_paper_name())
        if book.get_orientation() is not None:  # 0 is legal
            self.book.set_orientation(book.get_orientation())
        if book.get_paper_metric() is not None:  # 0 is legal
            self.book.set_paper_metric(book.get_paper_metric())
        if book.get_custom_paper_size():
            self.book.set_custom_paper_size(book.get_custom_paper_size())
        if book.get_margins():
            self.book.set_margins(book.get_margins())
        if book.get_format_name():
            self.book.set_format_name(book.get_format_name())
        if book.get_output():
            self.book.set_output(book.get_output())
        if book.get_dbname() != self._db.get_save_path():
            WarningDialog(
                _("Different database"),
                _(
                    "This book was created with the references to database "
                    "%s.\n\n This makes references to the central person "
                    "saved in the book invalid.\n\n"
                    "Therefore, the central person for each item is being set "
                    "to the active person of the currently opened database."
                )
                % book.get_dbname(),
                parent=self.window,
            )

        self.book.clear()
        self.book_model.clear()
        for saved_item in book.get_item_list():
            name = saved_item.get_name()
            item = BookItem(self._db, name)

            # The option values were loaded magically by the book parser.
            # But they still need to be applied to the menu options.
            opt_dict = item.option_class.handler.options_dict
            orig_opt_dict = saved_item.option_class.handler.options_dict
            menu = item.option_class.menu
            for optname in opt_dict:
                opt_dict[optname] = orig_opt_dict[optname]
                menu_option = menu.get_option_by_name(optname)
                if menu_option:
                    menu_option.set_value(opt_dict[optname])

            _initialize_options(item.option_class, self.dbstate, self.uistate)
            item.set_style_name(saved_item.get_style_name())
            self.book.append_item(item)

            data = [item.get_translated_name(), item.get_category(), item.get_name()]

            data[2] = item.option_class.get_subject()
            self.book_model.add(data)

    def on_add_clicked(self, obj):
        """
        Add an item to the current selections.

        Use the selected available item to get the item's name in the registry.
        """
        store, the_iter = self.avail_model.get_selected()
        if not the_iter:
            return
        data = self.avail_model.get_data(the_iter, list(range(self.avail_nr_cols)))
        item = BookItem(self._db, data[2])
        _initialize_options(item.option_class, self.dbstate, self.uistate)
        data[2] = item.option_class.get_subject()
        self.book_model.add(data)
        self.book.append_item(item)

    def on_remove_clicked(self, obj):
        """
        Remove the item from the current list of selections.
        """
        store, the_iter = self.book_model.get_selected()
        if not the_iter:
            return
        row = self.book_model.get_selected_row()
        self.book.pop_item(row)
        self.book_model.remove(the_iter)

    def on_clear_clicked(self, obj):
        """
        Clear the whole current book.
        """
        self.book_model.clear()
        self.book.clear()

    def on_up_clicked(self, obj):
        """
        Move the currently selected item one row up in the selection list.
        """
        row = self.book_model.get_selected_row()
        if not row or row == -1:
            return
        store, the_iter = self.book_model.get_selected()
        data = self.book_model.get_data(the_iter, list(range(self.book_nr_cols)))
        self.book_model.remove(the_iter)
        self.book_model.insert(row - 1, data, None, 1)
        item = self.book.pop_item(row)
        self.book.insert_item(row - 1, item)

    def on_down_clicked(self, obj):
        """
        Move the currently selected item one row down in the selection list.
        """
        row = self.book_model.get_selected_row()
        if row + 1 >= self.book_model.count or row == -1:
            return
        store, the_iter = self.book_model.get_selected()
        data = self.book_model.get_data(the_iter, list(range(self.book_nr_cols)))
        self.book_model.remove(the_iter)
        self.book_model.insert(row + 1, data, None, 1)
        item = self.book.pop_item(row)
        self.book.insert_item(row + 1, item)

    def on_setup_clicked(self, obj):
        """
        Configure currently selected item.
        """
        store, the_iter = self.book_model.get_selected()
        if not the_iter:
            WarningDialog(
                _("No selected book item"),
                _("Please select a book item to configure."),
                parent=self.window,
            )
            return
        row = self.book_model.get_selected_row()
        item = self.book.get_item(row)
        option_class = item.option_class
        option_class.handler.set_default_stylesheet_name(item.get_style_name())
        item.is_from_saved_book = bool(self.book.get_name())
        item_dialog = BookItemDialog(self.dbstate, self.uistate, item, self.track)

        while True:
            response = item_dialog.window.run()
            if response == Gtk.ResponseType.OK:
                # dialog will be closed by connect, now continue work while
                # rest of dialog is unresponsive, release when finished
                style = option_class.handler.get_default_stylesheet_name()
                item.set_style_name(style)
                subject = option_class.get_subject()
                self.book_model.model.set_value(the_iter, 2, subject)
                self.book.set_item(row, item)
                item_dialog.close()
                break
            elif response == Gtk.ResponseType.CANCEL:
                item_dialog.close()
                break
            elif response == Gtk.ResponseType.DELETE_EVENT:
                # just stop, in ManagedWindow, delete-event is already coupled to
                # correct action.
                break
        opt_dict = option_class.handler.options_dict
        for optname in opt_dict:
            menu_option = option_class.menu.get_option_by_name(optname)
            if menu_option:
                menu_option.set_value(opt_dict[optname])

    def book_button_press(self, obj, event):
        """
        Double-click on the current book selection is the same as setup.
        Right click evokes the context menu.
        """
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            self.on_setup_clicked(obj)
        elif is_right_click(event):
            self.build_book_context_menu(event)

    def avail_button_press(self, obj, event):
        """
        Double-click on the available selection is the same as add.
        Right click evokes the context menu.
        """
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            self.on_add_clicked(obj)
        elif is_right_click(event):
            self.build_avail_context_menu(event)

    def build_book_context_menu(self, event):
        """Builds the menu with item-centered and book-centered options."""

        store, the_iter = self.book_model.get_selected()
        if the_iter:
            sensitivity = 1
        else:
            sensitivity = 0
        entries = [
            (_("_Up"), self.on_up_clicked, sensitivity),
            (_("_Down"), self.on_down_clicked, sensitivity),
            (_("Setup"), self.on_setup_clicked, sensitivity),
            (_("_Remove"), self.on_remove_clicked, sensitivity),
            ("", None, 0),
            (_("Clear the book"), self.on_clear_clicked, 1),
            (_("_Save"), self.on_save_clicked, 1),
            (_("_Open"), self.on_open_clicked, 1),
            (_("_Edit"), self.on_edit_clicked, 1),
        ]

        self.menu1 = Gtk.Menu()  # TODO could this be just a local "menu ="?
        self.menu1.set_reserve_toggle_size(False)
        for title, callback, sensitivity in entries:
            item = Gtk.MenuItem.new_with_mnemonic(title)
            Gtk.Label.new_with_mnemonic
            if callback:
                item.connect("activate", callback)
            else:
                item = Gtk.SeparatorMenuItem()
            item.set_sensitive(sensitivity)
            item.show()
            self.menu1.append(item)
        self.menu1.popup_at_pointer(event)

    def build_avail_context_menu(self, event):
        """Builds the menu with the single Add option."""

        store, the_iter = self.avail_model.get_selected()
        if the_iter:
            sensitivity = 1
        else:
            sensitivity = 0
        entries = [
            (_("_Add"), self.on_add_clicked, sensitivity),
        ]

        self.menu2 = Gtk.Menu()  # TODO could this be just a local "menu ="?
        self.menu2.set_reserve_toggle_size(False)
        for title, callback, sensitivity in entries:
            item = Gtk.MenuItem.new_with_mnemonic(title)
            if callback:
                item.connect("activate", callback)
            item.set_sensitive(sensitivity)
            item.show()
            self.menu2.append(item)
        self.menu2.popup_at_pointer(event)

    def on_close_clicked(self, obj):
        """
        close the BookSelector dialog, saving any changes if needed
        """
        if self.book_list.get_needs_saving():
            self.book_list.save()
        ManagedWindow.close(self, *obj)

    def on_book_ok_clicked(self, obj):
        """
        Run final BookDialog with the current book.
        """
        if self.book.get_item_list():
            old_paper_name = self.book.get_paper_name()  # from books.xml
            old_orientation = self.book.get_orientation()
            old_paper_metric = self.book.get_paper_metric()
            old_custom_paper_size = self.book.get_custom_paper_size()
            old_margins = self.book.get_margins()
            old_format_name = self.book.get_format_name()
            old_output = self.book.get_output()
            BookDialog(
                self.dbstate, self.uistate, self.book, BookOptions, track=self.track
            )
            new_paper_name = self.book.get_paper_name()
            new_orientation = self.book.get_orientation()
            new_paper_metric = self.book.get_paper_metric()
            new_custom_paper_size = self.book.get_custom_paper_size()
            new_margins = self.book.get_margins()
            new_format_name = self.book.get_format_name()
            new_output = self.book.get_output()
            # only books in the booklist have a name (not "ad hoc" ones)
            if self.book.get_name() and (
                old_paper_name != new_paper_name
                or old_orientation != new_orientation
                or old_paper_metric != new_paper_metric
                or old_custom_paper_size != new_custom_paper_size
                or old_margins != new_margins
                or old_format_name != new_format_name
                or old_output != new_output
            ):
                self.book.set_dbname(self._db.get_save_path())
                self.book_list.set_book(self.book.get_name(), self.book)
                self.book_list.set_needs_saving(True)
            if self.book_list.get_needs_saving():
                self.book_list.save()
        else:
            WarningDialog(
                _("No items"), _("This book has no items."), parent=self.window
            )
            return
        self.close()

    def on_save_clicked(self, obj):
        """
        Save the current book in the xml booklist file.
        """
        if not self.book.get_item_list():
            WarningDialog(
                _("No items"), _("This book has no items."), parent=self.window
            )
            return
        name = str(self.name_entry.get_text())
        if not name:
            WarningDialog(
                _("No book name"),
                _(
                    "You are about to save away a book with no name.\n\n"
                    "Please give it a name before saving it away."
                ),
                parent=self.window,
            )
            return
        if name in self.book_list.get_book_names():
            qqq = QuestionDialog2(
                _("Book name already exists"),
                _(
                    "You are about to save away a "
                    "book with a name which already exists."
                ),
                _("Proceed"),
                _("Cancel"),
                parent=self.window,
            )
            if not qqq.run():
                return

        # previously, the same book could be added to the booklist
        # under multiple names, which became different books once the
        # booklist was saved into a file so everything was fine, but
        # this created a problem once the paper settings were added
        # to the Book object in the BookDialog, since those settings
        # were retrieved from the Book object in BookList.save, so mutiple
        # books (differentiated by their names) were assigned the
        # same paper values, so the solution is to make each Book be
        # unique in the booklist, so if multiple copies are saved away
        # only the last one will get the paper values assigned to it
        # (although when the earlier books are then eventually run,
        # they'll be assigned paper values also)
        self.book.set_name(name)
        self.book.set_dbname(self._db.get_save_path())
        self.book_list.set_book(name, self.book)
        self.book_list.set_needs_saving(True)  # user clicked on save
        self.book = Book(self.book, exact_copy=False)  # regenerate old items
        self.book.set_name(name)
        self.book.set_dbname(self._db.get_save_path())

    def on_open_clicked(self, obj):
        """
        Run the BookListDisplay dialog to present the choice of books to open.
        """
        booklistdisplay = BookListDisplay(
            self.book_list, nodelete=True, dosave=False, parent=self.window
        )
        booklistdisplay.top.destroy()
        book = booklistdisplay.selection
        if book:
            self.open_book(book)
            self.name_entry.set_text(book.get_name())
            self.book.set_name(book.get_name())

    def on_edit_clicked(self, obj):
        """
        Run the BookListDisplay dialog to present the choice of books to delete.
        """
        booklistdisplay = BookListDisplay(
            self.book_list, nodelete=False, dosave=True, parent=self.window
        )
        booklistdisplay.top.destroy()
        book = booklistdisplay.selection
        if book:
            self.open_book(book)
            self.name_entry.set_text(book.get_name())
            self.book.set_name(book.get_name())


# ------------------------------------------------------------------------
#
# Book Item Options dialog
#
# ------------------------------------------------------------------------
class BookItemDialog(ReportDialog):
    """
    This class overrides the interface methods common for different reports
    in a way specific for this report. This is a book item dialog.
    """

    def __init__(self, dbstate, uistate, item, track=[]):
        option_class = item.option_class
        name = item.get_name()
        translated_name = item.get_translated_name()
        self.category = CATEGORY_BOOK
        self.database = dbstate.db
        self.option_class = option_class
        self.is_from_saved_book = item.is_from_saved_book
        ReportDialog.__init__(
            self, dbstate, uistate, option_class, name, translated_name, track
        )

    def on_ok_clicked(self, obj):
        """The user is satisfied with the dialog choices. Parse all options
        and close the window."""

        # Preparation
        self.parse_style_frame()
        self.parse_user_options()

        self.options.handler.save_options()

    def setup_target_frame(self):
        """Target frame is not used."""
        pass

    def parse_target_frame(self):
        """Target frame is not used."""
        return 1

    def init_options(self, option_class):
        try:
            if issubclass(option_class, object):
                self.options = option_class(self.raw_name, self.database)
        except TypeError:
            self.options = option_class
        if not self.is_from_saved_book:
            self.options.load_previous_values()

    def add_user_options(self):
        """
        Generic method to add user options to the gui.
        """
        if not hasattr(self.options, "menu"):
            return
        menu = self.options.menu
        options_dict = self.options.options_dict
        for category in menu.get_categories():
            for name in menu.get_option_names(category):
                option = menu.get_option(category, name)

                # override option default with xml-saved value:
                if name in options_dict:
                    option.set_value(options_dict[name])

                widget, label = make_gui_option(
                    option,
                    self.dbstate,
                    self.uistate,
                    self.track,
                    self.is_from_saved_book,
                )
                if widget is not None:
                    if label:
                        self.add_frame_option(category, option.get_label(), widget)
                    else:
                        self.add_frame_option(category, "", widget)


# -------------------------------------------------------------------------
#
# _BookFormatComboBox
#
# -------------------------------------------------------------------------
class _BookFormatComboBox(Gtk.ComboBox):
    """
    Build a menu of report types that are appropriate for a book
    """

    def __init__(self, active):
        Gtk.ComboBox.__init__(self)

        pmgr = GuiPluginManager.get_instance()
        self.__bookdoc_plugins = []
        for plugin in pmgr.get_docgen_plugins():
            if plugin.get_text_support() and plugin.get_draw_support():
                self.__bookdoc_plugins.append(plugin)

        self.store = Gtk.ListStore(GObject.TYPE_STRING)
        self.set_model(self.store)
        cell = Gtk.CellRendererText()
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)

        index = 0
        active_index = 0
        for plugin in self.__bookdoc_plugins:
            name = plugin.get_name()
            self.store.append(row=[name])
            if plugin.get_extension() == active:
                active_index = index
            index += 1
        self.set_active(active_index)

    def get_active_plugin(self):
        """
        Get the plugin represented by the currently active selection.
        """
        return self.__bookdoc_plugins[self.get_active()]


# ------------------------------------------------------------------------
#
# The final dialog - paper, format, target, etc.
#
# ------------------------------------------------------------------------
class BookDialog(DocReportDialog):
    """
    A usual Report.Dialog subclass.

    Create a dialog selecting target, format, and paper/HTML options.
    """

    def __init__(self, dbstate, uistate, book, options, track=[]):
        self.format_menu = None
        self.options = options
        self.page_html_added = False
        self.book = book
        self.title = _("Generate Book")
        self.database = dbstate.db
        DocReportDialog.__init__(
            self, dbstate, uistate, options, "book", self.title, track=track
        )
        self.options.options_dict["bookname"] = self.book.get_name()

        while True:
            response = self.window.run()
            if response != Gtk.ResponseType.HELP:
                break
        if response == Gtk.ResponseType.OK:
            handler = self.options.handler
            if self.book.get_paper_name() != handler.get_paper_name():
                self.book.set_paper_name(handler.get_paper_name())
            if self.book.get_orientation() != handler.get_orientation():
                self.book.set_orientation(handler.get_orientation())
            if self.book.get_paper_metric() != handler.get_paper_metric():
                self.book.set_paper_metric(handler.get_paper_metric())
            if self.book.get_custom_paper_size() != handler.get_custom_paper_size():
                self.book.set_custom_paper_size(handler.get_custom_paper_size())
            if self.book.get_margins() != handler.get_margins():
                self.book.set_margins(handler.get_margins())
            if self.book.get_format_name() != handler.get_format_name():
                self.book.set_format_name(handler.get_format_name())
            if self.book.get_output() != self.options.get_output():
                self.book.set_output(self.options.get_output())
            try:
                self.make_book()
            except (IOError, OSError) as msg:
                ErrorDialog(str(msg), parent=self.window)
        if response != Gtk.ResponseType.DELETE_EVENT:  # already closed
            self.close()

    def setup_style_frame(self):
        pass

    def setup_other_frames(self):
        pass

    def parse_style_frame(self):
        pass

    def get_title(self):
        """get the title"""
        return self.title

    def get_header(self, name):
        """get the header"""
        return _("Gramps Book")

    def make_doc_menu(self, active=None):
        """Build a menu of document types that are appropriate for
        this text report.  This menu will be generated based upon
        whether the document requires table support, etc."""
        self.format_menu = _BookFormatComboBox(active)

    def on_help_clicked(self, *obj):
        display_help(WIKI_HELP_PAGE, GENERATE_WIKI_HELP_SEC)

    def make_document(self):
        """Create a document of the type requested by the user."""
        user = User(uistate=self.uistate)
        self.rptlist = []
        selected_style = StyleSheet()

        pstyle = self.paper_frame.get_paper_style()
        self.doc = self.format(None, pstyle, uistate=self.uistate)

        for item in self.book.get_item_list():
            item.option_class.set_document(self.doc)
            report_class = item.get_write_item()
            obj = (
                write_book_item(self.database, report_class, item.option_class, user),
                item.get_translated_name(),
            )
            self.rptlist.append(obj)
            append_styles(selected_style, item)

        self.doc.set_style_sheet(selected_style)
        self.doc.open(self.target_path)

    def make_book(self):
        """
        The actual book. Start it out, then go through the item list
        and call each item's write_book_item method (which were loaded
        by the previous make_document method).
        """

        try:
            self.doc.init()
            newpage = 0
            for rpt, name in self.rptlist:
                if newpage:
                    self.doc.page_break()
                newpage = 1
                if rpt:
                    rpt.begin_report()
                    rpt.write_report()
            self.doc.close()
        except ReportError as msg:
            (msg1, msg2) = msg.messages()
            msg2 += " (%s)" % name  # which report has the error?
            ErrorDialog(msg1, msg2, parent=self.uistate.window)
            return
        except FilterError as msg:
            (msg1, msg2) = msg.messages()
            ErrorDialog(msg1, msg2, parent=self.uistate.window)
            return

        if self.open_with_app.get_active():
            open_file_with_default_application(self.target_path, self.uistate)

    def init_options(self, option_class):
        try:
            if issubclass(option_class, object):
                self.options = option_class(self.raw_name, self.database)
        except TypeError:
            self.options = option_class
        self.options.load_previous_values()
        handler = self.options.handler
        if self.book.get_paper_name():
            handler.set_paper_name(self.book.get_paper_name())
        if self.book.get_orientation() is not None:  # 0 is legal
            handler.set_orientation(self.book.get_orientation())
        if self.book.get_paper_metric() is not None:  # 0 is legal
            handler.set_paper_metric(self.book.get_paper_metric())
        if self.book.get_custom_paper_size():
            handler.set_custom_paper_size(self.book.get_custom_paper_size())
        if self.book.get_margins():
            handler.set_margins(self.book.get_margins())
        if self.book.get_format_name():
            handler.set_format_name(self.book.get_format_name())
        if self.book.get_output():
            self.options.set_output(self.book.get_output())


# ------------------------------------------------------------------------
#
# Generic task function for book
#
# ------------------------------------------------------------------------
def write_book_item(database, report_class, options, user):
    """
    Write the report using options set.
    All user dialog has already been handled and the output file opened.
    """
    try:
        return report_class(database, options, user)
    except ReportError as msg:
        (msg1, msg2) = msg.messages()
        ErrorDialog(msg1, msg2, parent=user.uistate.window)
    except FilterError as msg:
        (msg1, msg2) = msg.messages()
        ErrorDialog(msg1, msg2, parent=user.uistate.window)
    except:
        LOG.error("Failed to write book item.", exc_info=True)
    return None



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013 Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Unittest for editreference.py"""

import unittest
from unittest.mock import Mock, patch
import os

from gramps.gen.lib import (
    Person,
    Family,
    Event,
    Source,
    Place,
    Citation,
    Repository,
    Media,
    Note,
    Tag,
)
from gramps.cli.user import User
from gramps.gen.dbstate import DbState
from gramps.gen.db.utils import make_database
from gramps.gen.db import DbTxn
from gramps.gui.editors.editreference import EditReference


class MockWindow:
    def set_transient_for(self, *args, **kwargs):
        pass

    def show_all(self):
        pass


class MockEditReference(EditReference):
    def __init__(self, dbstate, uistate, track, source, source_ref, update):
        self.window = MockWindow()
        super().__init__(dbstate, uistate, track, source, source_ref, update)


class TestEditReference(unittest.TestCase):
    def test_editreference(self):
        dbstate = DbState()
        db = make_database("sqlite")
        path = "/tmp/edit_ref_test"
        try:
            os.mkdir(path)
        except:
            pass
        db.load(path)
        dbstate.change_database(db)
        source = Place()
        source.gramps_id = "P0001"
        with DbTxn("test place", dbstate.db) as trans:
            dbstate.db.add_place(source, trans)
        editor = MockEditReference(
            dbstate, uistate=None, track=[], source=source, source_ref=None, update=None
        )
        with patch("gramps.gui.editors.editreference.ErrorDialog") as MockED:
            editor.check_for_duplicate_id("Place")
            self.assertTrue(MockED.called)


if __name__ == "__main__":
    unittest.main()



"""
WithinArea : used to verify if a place is contained in a specific area
"""

#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2015       Nick Hall
# Copyright (C) 2017-      Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------

from math import pi, cos, hypot
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....errors import FilterError
from ....const import GRAMPS_LOCALE as glocale
from .. import Rule
from ....utils.place import conv_lat_lon


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Union
from ....lib import Place
from ....db import Database


_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# WithinArea
#
# -------------------------------------------------------------------------
class WithinArea(Rule):
    """
    Rule that checks for a place within an area
    """

    labels = [_("ID:"), _("Value:"), _("Units:")]
    name = _("Places within an area")
    description = _("Matches places within a given distance of another place")
    category = _("Position filters")
    handle = None
    radius: float = 0.0
    latitude: Union[float, None] = None
    longitude: Union[float, None] = None

    def prepare(self, db: Database, user):
        ref_place = db.get_place_from_gramps_id(self.list[0])
        self.handle = None
        self.latitude = None
        self.longitude = None
        if ref_place:
            self.handle = ref_place.handle
            latitude = ref_place.lat
            if latitude == "":
                latitude = None
                return
            longitude = ref_place.long
            self.latitude, self.longitude = conv_lat_lon(latitude, longitude, "D.D8")  # type: ignore
            if self.latitude is None or self.longitude is None:
                raise FilterError(
                    _("Cannot use the filter 'within area'"),
                    _(
                        "The place you selected contains bad"
                        " coordinates. Please, run the tool "
                        "'clean input data'"
                    ),
                )

            val = self.list[1]
            if isinstance(val, str):
                val = re.sub(r"\D", "", val)  # suppress all alpha characters
            value = int(val)
            unit = int(self.list[2])
            # earth perimeter in kilometers for latitude
            # 2 * pi * (6371 * cos(latitude/180*pi))
            # so 1 degree correspond to the result above / 360
            earth_perimeter = 2 * pi * (6371 * cos(float(self.latitude) / 180 * pi))
            if unit == 0:  # kilometers
                self.radius = float(value / (earth_perimeter / 360))
            elif unit == 1:  # miles
                self.radius = float((value / (earth_perimeter / 360)) / 0.62138)
            else:  # degrees
                self.radius = float(value)
            self.radius = self.radius / 2

    def apply_to_one(self, db: Database, place: Place) -> bool:
        latit: Union[float, None] = None
        longit: Union[float, None] = None

        if not (place and self.handle and self.latitude and self.longitude):
            return False
        if place:
            lat = place.lat
            lon = place.long
        if lat and lon:
            latit, longit = conv_lat_lon(lat, lon, "D.D8")  # type: ignore
            return (
                latit  # type: ignore
                and longit
                and (
                    hypot(
                        float(self.latitude) - float(latit),
                        float(self.longitude) - float(longit),
                    )
                    <= self.radius
                )
            )
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013    Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....lib.childreftype import ChildRefType
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Family
from ....db import Database


# -------------------------------------------------------------------------
#
# HasTwins
#
# -------------------------------------------------------------------------
class HasTwins(Rule):
    """Rule that checks for a family with twins"""

    name = _("Families with twins")
    description = _("Matches families with twins")
    category = _("Child filters")

    def apply_to_one(self, db: Database, family: Family) -> bool:
        date_list = []
        for childref in family.child_ref_list:
            if int(childref.mrel.value) == ChildRefType.BIRTH:
                child = db.get_person_from_handle(childref.ref)
                if 0 <= child.birth_ref_index < len(child.event_ref_list):
                    birthref = child.event_ref_list[child.birth_ref_index]
                    if birthref:
                        birth = db.get_event_from_handle(birthref.ref)
                        sortval = birth.date.sortval
                        if sortval != 0:
                            if sortval in date_list:
                                return True
                            else:
                                date_list.append(sortval)
        return False
